# -*- coding: utf8 -*-
import copy
import six
from FGAme.events import EventDispatcher, EventDispatcherMeta, signal
from FGAme.mathtools import Vec2, asvector, sin, cos, null2D, shapes
from FGAme.physics.flags import BodyFlags as flags
from FGAme.physics.forces import ForceProperty
from FGAme.physics.bodies.body_interfaces import (HasAABB, HasGlobalForces,
                                                  HasInertia)

__all__ = ['Body', 'LinearRigidBody']


###############################################################################
#                              Funções úteis
###############################################################################
NOT_IMPLEMENTED = NotImplementedError('must be implemented at child classes')
INERTIA_SCALE = 1
INF = float('inf')


def flag_property(flag):
    '''Retorna uma propriedade que controla a flag especificada no argumento'''

    def fget(self):
        return self.flags & flag

    def fset(self, value):
        if value:
            self.flags |= flag
        else:
            self.flags &= ~flag
    return property(fget, fset)

# TODO: mover para FGAme.util


def do_nothing(*args, **kwds):
    '''A handle that does nothing'''


def raises_method(exception=NOT_IMPLEMENTED):
    '''Returns a method that raises the given exception'''

    def method(self, *args, **kwds):
        raise exception
    return method


###############################################################################
#          Classes Base --- todos objetos físicos derivam delas
###############################################################################
Body = None


class PhysElementMeta(EventDispatcherMeta):

    '''Metaclasse para todas as classes que representam objetos físicos'''

    BLACKLIST = ['_is_mixin_', '__module__', '__doc__', '__module__',
                 '__dict__', '__weakref__', '__slots__', '__subclasshook__']

    def __new__(cls, name, bases, ns):
        # Aplica transformações da EventDispatcherMeta
        has_slots = '__slots__' in ns
        slots = ns.setdefault('__slots__', [])
        ns = cls._populate_namespace(name, bases, ns)

        # Remove __slots__ creiados pela EventDispatcher
        ns['__slots__'] = slots

        # Força EventDispatcher ser a classe mãe na hierarquia
        if bases == (object,):
            bases = (EventDispatcher,)

        # Insere uma cláusula de __slots__ vazia
        ns.setdefault('__slots__', [])

        # Verifica se existem classes mix-in entre as bases
        true_bases = tuple(B for B in bases
                           if not getattr(B, '_is_mixin_', False))
        if len(bases) != len(true_bases):
            for i, B in enumerate(bases):
                if not getattr(B, '_is_mixin_', False):
                    continue

                # Encontra quais variáveis/métodos foram definidos nas classes
                # mixin
                allvars = {attr: getattr(B, attr) for attr in dir(B)}
                for var in cls.BLACKLIST:
                    if var in allvars:
                        del allvars[var]
                for attr, value in list(allvars.items()):
                    if hasattr(object, attr):
                        if getattr(object, attr) is value:
                            del allvars[attr]

                # Insere as variáveis que não foram definidas em ns nem em
                # nenhuma base anterior
                prev_bases = bases[:i]
                for attr in list(allvars.keys()):
                    if attr in ns:
                        del allvars[attr]
                        continue

                    for B in prev_bases:
                        if hasattr(B, attr):
                            del allvars[attr]
                            break

                # Em Python 2, utiliza o objeto im_func dos métodos, caso ele
                # exista
                allvars = {name: getattr(var, 'im_func', var) for (name, var)
                           in allvars.items()}

                ns.update(allvars)

                # Atualiza a lista de __slots__ utilizando o atributo _slots_
                # do mixin e toda hierarquia inferior
                for tt in B.mro():
                    slots = getattr(tt, '_slots_', [])
                    ns['__slots__'].extend(slots)

        # Atualiza a lista de propriedades e retira slots repetidos
        ns['__slots__'] = list(set(ns['__slots__']))
        if not has_slots:
            del ns['__slots__']
        else:
            if Body is None:
                ns['__slots__'].append('__dict__')
        new = type.__new__(cls, name, true_bases, ns)

        # Atualiza as docstrings vazias utilizando as docstrings da primeira
        # base válida
        mro = new.mro()[1:-1]
        for name, method in ns.items():
            if not callable(method):
                continue

            doc = getattr(method, '__doc__', '<no docstring>')
            if not doc:
                for base in mro:
                    try:
                        other = getattr(base, name)
                    except AttributeError:
                        continue
                    if other.__doc__:
                        # TODO: six it
                        try:
                            method.__doc__ = other.__doc__
                        except AttributeError:
                            pass
                    break

        return new


@six.add_metaclass(PhysElementMeta)
class Body(EventDispatcher, HasAABB, HasGlobalForces, HasInertia):

    '''Classe mãe de todos objetos com propriedades dinâmicas.

    Attributes
    ----------

    ::
        **Propriedades físicas do objeto**
    mass
        Massa do objeto. Possui o valor padrão de 1. Uma massa infinita
        transforma o objeto num objeto cinemático que não responde a forças
        lineares.

    ::
        **Variáveis dinâmicas**
    pos
        Posição do centro de massa do objeto
    vel
        Velocidade linear medida a partir do centro de massa
    accel
        Aceleração acumulada recalculada em cada frame

    ::
         **Forças locais**

    gravity
        Valor da aceleração da gravidade aplicada ao objeto
    damping, adamping
        Constante de amortecimento linear/angular para forças viscosas
        aplicadas ao objeto
    owns_gravity, owns_damping, owns_adamping
        Se Falso (padrão) utiliza os valores de gravity e damping/adamping
        fornecidos pelo mundo
    accel_static
        Caso verdadeiro, aplica as acelerações de gravidade, damping/adamping
        no objeto mesmo se ele for estático

    ::
        **Propriedades físicas do objeto**
    inertia
        Momento de inércia do objeto com relação ao eixo z no centro de massa.
        Calculado automaticamente a partir da geometria e densidade do objeto.
        Caso seja infinito, o objeto não responderá a torques.
    ROG, ROG_sqr
        Raio de giração e o quadrado do raio de giração. Utilizado para
        calcular o momento de inércia: $I = M R^2$, onde I é o momento de
        inércia, M a massa e R o raio de giração.
    density
        Densidade de massa: massa / área
    area
        Área que o objeto ocupa

    ::
        **Variáveis dinâmicas**
    theta
        Ângulo da rotação em torno do eixo saindo do centro de massa do objeto
    omega
        Velocidade angular de rotação

    ::
        **Caixa de contorno**
    xmin, xmax, ymin, ymax
        Limites da caixa de contorno alinhada aos eixos que envolve o objeto
    bbox
        Uma tupla com (xmin, xmax, ymin, ymax)
    shape
        Uma tupla (Lx, Ly) com a forma caixa de contorno nos eixos x e y.
    rect
        Uma tupla com (xmin, ymin, Lx, Ly)

        **Flags de colisão**
    col_group : int ou sequência
        Número inteiro positivoque representa o grupo a qual o objeto pertence.
        Objetos do mesmo grupo nunca colidem entre si. `col_group` pode ser uma
        sequência de valores caso o objeto pertença a vários grupos. O grupo
        zero (padrão) é tratado de forma especial e todos os objetos deste
        grupo colidem entre si.
    col_layer : int ou sequência
        Número inteiro positivo representando a camdada à qual o objeto
        pertence. O valor padrão é zero. Apenas objetos que estão na mesma
        camada colidem entre si. `col_layer` pode ser uma sequência de valores
        caso o objeto participe de várias camadas.
    '''

    __slots__ = [
        'flags', 'cbb_radius', '_baseshape', '_shape', '_aabb',
        '_pos', '_vel', '_accel', '_theta', '_omega', '_alpha',
        '_invmass', '_invinertia', '_e_vel', '_e_omega', '_world',
        '_col_layer', '_col_group_mask',
    ]

    DEFAULT_FLAGS = 0 | flags.can_rotate | flags.dirty_shape | flags.dirty_aabb

    def __init__(self, pos=null2D, vel=null2D, theta=0.0, omega=0.0,
                 mass=None, density=None, inertia=None,
                 gravity=None, damping=None, adamping=None,
                 restitution=None, friction=None,
                 baseshape=None, world=None, col_layer=0, col_group=0,
                 flags=DEFAULT_FLAGS):

        self._world = world
        EventDispatcher.__init__(self)

        # Flags de objeto
        self.flags = flags

        # Variáveis de estado #################################################
        self._pos = asvector(pos)
        self._vel = asvector(vel)
        self._e_vel = null2D
        self._e_omega = 0.0
        self._theta = float(theta)
        self._omega = float(omega)
        self._accel = null2D
        self._alpha = 0.0

        # Harmoniza massa, inércia e densidade ################################
        self._baseshape = self._shape = baseshape
        self._aabb = getattr(baseshape, 'aabb', None)
        if density is not None:
            density = float(density)
            if mass is None:
                mass = density * self.area()
            else:
                mass = float(mass)
            if inertia is None:
                inertia = density * \
                    self.area() * self.ROG_sqr() / INERTIA_SCALE
            else:
                inertia = float(inertia)

        elif mass is not None:
            mass = float(mass)
            try:
                density = mass / self.area()
            except ZeroDivisionError:
                density = float('inf')
            if inertia is None:
                inertia = mass * self.ROG_sqr() / INERTIA_SCALE
            else:
                inertia = float(inertia)

        else:
            density = 1.0
            mass = density * self.area()
            if inertia is None:
                inertia = density * \
                    self.area() * self.ROG_sqr() / INERTIA_SCALE
            else:
                inertia = float(inertia)

        self._invmass = 1.0 / mass
        self._invinertia = 1.0 / inertia
        self._density = float(density)

        # Controle de parâmetros físicos locais ###############################
        self._gravity = null2D
        self._damping = self._adamping = 0.0
        self._friction = 0.0
        self._restitution = 1.0
        if damping is not None:
            self.damping = damping
        if adamping is not None:
            self.adamping = adamping
        if gravity is not None:
            self.gravity = gravity
        if restitution is not None:
            self.restitution = restitution
        if friction is not None:
            self.friction = friction

        # Vínculos e contatos #################################################
        self._contacts = []
        self._joints = []

        # Filtros de colisões #################################################
        # Colide se objetos estão em groupos diferentes (exceto os que estão
        # no grupo 0) e no mesmo layer
        if col_layer:
            if isinstance(col_layer, int):
                self._col_layer_mask = 1 << col_layer
            else:
                mask = 0
                for n in col_layer:
                    mask |= 1 << n
                self._col_layer_mask = mask
        else:
            self._col_layer_mask = 0

        if col_group:
            if isinstance(col_group, int):
                self._col_group_mask = 1 << (col_group - 1)
            else:
                mask = 0
                for n in col_group:
                    mask |= 1 << (n - 1)
                self._col_group_mask = mask
        else:
            self._col_group_mask = 0

        # Presença em mundo ###################################################
        if world is not None:
            self._world.add(self)

    ###########################################################################
    #                            Serviços Python
    ###########################################################################

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return id(self)

    def __repr__(self):
        tname = type(self).__name__
        pos = ', '.join('%.1f' % x for x in self._pos)
        vel = ', '.join('%.1f' % x for x in self._vel)
        return '%s(pos=(%s), vel=(%s))' % (tname, pos, vel)

    # TODO: def __copy__(self): ?
    # TODO: def __getstate__(self): ?

    ###########################################################################
    #                      Propriedades geométricas
    ###########################################################################
    @property
    def cbb(self):
        '''Caixa de contorno circular que envolve o objeto'''

        return shapes.Circle(self.cbb_radius, self.pos)

    @property
    def aabb(self):
        '''Caixa de contorno alinhada aos eixos que envolve o objeto'''

        if self.flags & flags.dirty_aabb:
            self._aabb = self.bounding_box.get_aabb()
            self.flags &= flags.not_dirty_aabb
        return self._aabb

    @property
    def bounding_box(self):
        '''Retorna um objeto que representa o formato geométrico do corpo
        físico, ex.: Circle, AABB, Poly, etc'''

        if self.flags & flags.dirty_any:
            self._shape = self._baseshape.move(self.pos).rotate(self._theta)
            self.flags &= flags.not_dirty
        return self._shape

    # Propriedades da caixa de contorno #######################################

    ###########################################################################
    #                          Contatos e vínculos
    ###########################################################################
    def add_contact(self, contact):
        '''Registra um contato à lista de contatos do objeto'''

        L = self._contacts
        invmass = contact.A._invmass
        for i, C in enumerate(L):
            if invmass < C.A._invmass:
                L.insert(i, C)
                break
        else:
            L.append(contact)

    def remove_contact(self, contact):
        '''Remove um contato da lista de contatos do objeto'''

        try:
            self._contacts.remove(contact)
        except ValueError:
            pass

    ###########################################################################
    #                                Sinais
    ###########################################################################
    collision = signal('collision', num_args=1)
    frame_enter = signal('frame-enter')
    out_of_bounds = signal('out-of-bounds', num_args=1)

    ###########################################################################
    #                            Controle de flags
    ###########################################################################
    def set_flag(self, flag, value):
        '''Atribui um valor de verdade para a flag especificada.

        O argumento pode ser uma string ou a constante em FGAme.physics.flags
        associada à flag'''

        if isinstance(flag, str):
            flag = self._get_flag(flag)
        if value:
            self.flags |= flag
        else:
            self.flags &= ~flag

    def toggle_flag(self, flag):
        '''Inverte o valor da flag.

        O argumento pode ser uma string ou a constante em FGAme.physics.flags
        associada à flag'''

        if isinstance(flag, str):
            flag = self._get_flag(flag)

        value = not (self.flags & flag)
        if value:
            self.flags |= flag
        else:
            self.flags &= ~flag

    def get_flag(self, flag):
        '''Retorna o valor da flag.

        O argumento pode ser uma string ou a constante em FGAme.physics.flags
        associada à flag'''

        if isinstance(flag, str):
            flag = self._get_flag(flag)
        return bool(self.flags | flag)

    #
    # Propriedades definidas por flags
    #
    can_rotate = flag_property(flags.can_rotate)
    owns_gravity = flag_property(flags.owns_gravity)
    owns_damping = flag_property(flags.owns_damping)
    owns_adamping = flag_property(flags.owns_adamping)
    owns_restitution = flag_property(flags.owns_restitution)
    owns_friction = flag_property(flags.owns_friction)

    ###########################################################################
    #        Controle de criação e destruição e interação com o mundo
    ###########################################################################
    @property
    def world(self):
        if self._world is None:
            raise ValueError('trying to access rogue object')
        else:
            return self._world

    def set_active(self, value):
        '''Determina o estado ativo do objeto'''

        if value:
            self.flags &= flags.not_active
        else:
            self.world.set_active(self)
            self.flags |= flags.active

    def is_rogue(self):
        '''Retorna True se o objeto não estiver associado a uma simulação'''

        return (self._world is None)

    def destroy(self):
        '''Destrói o objeto físico'''

        if not self.is_rogue():
            if self in self.world:
                self.world.remove(self)

        # TODO: desaloca todos os sinais

    def copy(self):
        '''Cria uma cópia do objeto'''

        try:
            world, self._world = self._world, None
            cp = copy.copy(self)
        finally:
            self._world = world
        return cp

    ###########################################################################
    #                          Estado dinâmico
    ###########################################################################
    @property
    def omega(self):
        return self._omega

    @omega.setter
    def omega(self, value):
        if self.flags & flags.can_rotate:
            self._omega = value + 0.0
        elif value:
            self._raise_cannot_rotate_error()

    @property
    def theta(self):
        return self._theta

    @theta.setter
    def theta(self, value):
        if self.flags & flags.can_rotate:
            self._theta = value + 0.0
        elif value:
            self._raise_cannot_rotate_error()

    ###########################################################################
    #                        Propriedades físicas
    ###########################################################################
    force = ForceProperty()

    def linearK(self):
        '''Energia cinética das variáveis lineares'''

        if self._invmass:
            return self._vel.norm_sqr() / (2 * self._invmass)
        else:
            return 0.0

    def angularK(self):
        '''Energia cinética das variáveis angulares'''

        if self._invinertia:
            return self._omega ** 2 / (2 * self._invinertia)
        else:
            return 0.0

    def energyK(self):
        '''Energia cinética total'''

        return self.linearK() + self.angularK()

    def energyU(self):
        '''Energia potencial associada à gravidade e às forças externas'''

        return -self.gravity.dot(self._pos) / self._invmass

    def energy(self):
        '''Energia total do objeto'''

        return self.energyK() + self.energyU()

    def momentumP(self):
        '''Momentum linear'''

        try:
            return self._vel / self._invmass
        except ZeroDivisionError:
            return Vec2(float('inf'), float('inf'))

    def momentumL(self, pos_or_x=None, y=None):
        '''Momentum angular em torno do ponto dado (usa o centro de massa
        como padrão)

        Examples
        --------

        Considere uma partícula no ponto (0, 1) com velocidade (1, 0). É fácil
        calcular o momentum linear resultante

        >>> b1 = Body(pos=(0, 1), vel=(1, 0), mass=2)
        >>> b1.momentumL(0, 0)
        -2.0

        É lógico que este valor muda de acordo com o ponto de referência

        >>> b1.momentumL(0, 2)
        2.0

        Quando chamamos a função sem argumentos, o padrão é utilizar o centro
        de massa. Numa partícula pontual sem velocidade angular, este valor é
        sempre zero

        >>> b1.momentumL(b1.pos)
        0.0


        '''

        if pos_or_x is None:
            momentumL = 0.0
        elif y is None:
            delta_pos = self.pos - asvector(pos_or_x)
            momentumL = delta_pos.cross(self.momentumP())
        else:
            delta_pos = self.pos - Vec2(pos_or_x, y)
            momentumL = delta_pos.cross(self.momentumP())

        if self._invinertia:
            return momentumL + self.omega * self.inertia
        else:
            return momentumL

    ###########################################################################
    #                      Aplicação de forças e torques
    ###########################################################################
    def move(self, delta_or_x, y=None):
        '''Move o objeto por vetor de deslocamento delta'''

        if y is None:
            if delta_or_x is null2D:
                return
            self._pos += delta_or_x
        else:
            self._pos += (delta_or_x, y)

        self.flags |= flags.dirty_any

    def boost(self, delta_or_x, y=None):
        '''Adiciona um valor vetorial delta à velocidade linear'''

        if y is None:
            self._vel += delta_or_x
        else:
            self._vel += Vec2(delta_or_x, y)

    # Resposta a forças, impulsos e atualização da física #####################
    def init_accel(self):
        '''Inicializa o vetor de aceleração com os valores devidos à gravidade
        e ao amortecimento'''

        if self._damping:
            a = self._vel
            a *= -self._damping
            if self.gravity is not None:
                a += self._gravity
        elif self._gravity is not None:
            a = self._gravity
        else:
            a = null2D
        self._accel = a

    def apply_force(self, force, dt, pos=None, relative=False):
        '''Aplica uma força linear durante um intervalo de tempo dt'''

        if force is None:
            if self._invmass:
                self.apply_accel(self._accel, dt)
        else:
            self.apply_accel(force * self._invmass, dt)

        if pos is not None and self._invinertia:
            if relative:
                tau = pos.cross(force)
            else:
                tau = (pos - self._pos).cross(force)
            self.apply_torque(tau, dt)

    def apply_accel(self, a, dt, method=None):
        '''Aplica uma aceleração linear durante um intervalo de tempo dt.

        Tem efeito em objetos cinemáticos.

        Observations
        ------------

        Implementa a integração de Velocity-Verlet para o sistema. Este
        integrador é superior ao Euler por dois motivos: primeiro, trata-se de
        um integrador de ordem superior (O(dt^4) vs O(dt^2)). Além disto, ele
        possui a propriedade simplética, o que implica que o erro da energia
        não tende a divergir, mas sim oscilar ora positivamente ora
        negativamente em torno de zero. Isto é extremamente desejável para
        simulações de física que parecam realistas.

        A integração de Euler seria implementada como:

            x(t + dt) = x(t) + v(t) * dt
            v(t + dt) = v(t) + a(t) * dt

        Em código Python,

        >>> self.move(self.vel * dt + a * (dt**2/2))           # doctest: +SKIP
        >>> self.boost(a * dt)                                 # doctest: +SKIP

        Este método simples e intuitivo sofre com o efeito da "deriva de
        energia". Devido aos erros de aproximação, o valor da energia da
        solução numérica flutua com relação ao valor exato. Na grande maioria
        dos sistemas, esssa flutuação ocorre com mais probabilidade para a
        região de maior energia e portanto a energia tende a crescer
        continuamente, estragando a credibilidade da simulação.

        Velocity-Verlet está numa classe de métodos numéricos que não sofrem
        com esse problema. A principal desvantagem, no entanto, é que devemos
        manter uma variável adicional com o último valor conhecido da
        aceleração. Esta pequena complicação é mais que compensada pelo ganho
        em precisão numérica. O algorítmo consiste em:

            x(t + dt) = x(t) + v(t) * dt + a(t) * dt**2 / 2
            v(t + dt) = v(t) + [(a(t) + a(t + dt)) / 2] * dt

        O termo a(t + dt) normalemente só pode ser calculado se soubermos como
        obter as acelerações como função das posições x(t + dt). Na prática,
        cada iteração de .apply_accel() calcula o valor da posição em x(t + dt)
        e da velocidade no passo anterior v(t). Calcular v(t + dt) requer uma
        avaliação de a(t + dt), que só estará disponível na iteração seguinte.
        A próxima iteração segue então para calcular v(t + dt) e x(t + 2*dt), e
        assim sucessivamente.

        A ordem de acurácia de cada passo do algoritmo Velocity-Verlet é de
        O(dt^4) para uma força que dependa exclusivamente da posição e tempo.
        Caso haja dependência na velocidade, a acurácia reduz e ficaríamos
        sujeitos aos efeitos da deriva de energia. Normalmente as forças
        físicas que dependem da velocidade são dissipativas e tendem a reduzir
        a energia total do sistema muito mais rapidamente que a deriva de
        energia tende a fornecer energia espúria ao sistema. Deste modo, a
        acurácia ficaria reduzida, mas a simulação ainda manteria alguma
        credibilidade.
        '''
        a = asvector(a)

        if method is None or method == 'euler-semi-implicit':
            self.boost(a * dt)
            self.move(self._vel * dt + a * (0.5 * dt * dt))
        elif method == 'verlet':
            raise NotImplementedError
        elif method == 'euler':
            self.move(self._vel * dt + a * (0.5 * dt * dt))
            self.boost(a * dt)
        else:
            raise ValueError('invalid method: %r' % method)

    def apply_impulse(self, impulse_or_x, y=None, pos=None, relative=False):
        '''Aplica um impulso linear ao objeto. Isto altera sua velocidade
        linear com relação ao centro de massa.

        Se for chamado com dois agumentos aplica o impulso em um ponto
        específico e também resolve a dinâmica angular.

        Examples
        --------

        Considere um objeto criado na origem e outro na posição (4, 3)

        >>> b1 = Body(pos=(2, 0), mass=1)
        >>> b2 = Body(pos=(-1, 0), mass=2)

        Criamos um impulso J e aplicamos o mesmo em b1

        >>> J = Vec2(0, 2)
        >>> b1.apply_impulse(J)

        Note que isto afeta a velocidade do corpo de acordo com a fórmula
        $\delta v = J / m$:

        >>> b1.vel
        Vec(0, 2)

        Se aplicarmos um impulso oposto à b2, o resultado não deve alterar o
        momento linear, que permanece nulo

        >>> b2.apply_impulse(-J, pos=(0, 0)); b2.vel
        Vec(0, -1)
        >>> b1.momentumP() + b2.momentumP()
        Vec(0, 0)

        O exemplo anterior é bastante simples pois os dois objetos não possuem
        momentum de inércia. As mesmas leis de conservação valem na presença
        de rotações ou mesmo de velocidades iniciais

        >>> b1 = Body(pos=(0, 0), mass=1, inertia=1, vel=(2, 0), omega=1)
        >>> b2 = Body(pos=(4, 3), mass=2, inertia=2, vel=(-1, 1))

        Primeiro calculamos os momentos iniciais

        >>> P0 = b1.momentumP() + b2.momentumP()

        Note que o momentum linear exige um ponto de referência para a
        realização do cálculo. Escolhemos o centro de massa, mas os resultados
        devem valer para qualquer escolha arbitrária.

        >>> from FGAme.physics import center_of_mass
        >>> Rcm = center_of_mass(b1, b2)
        >>> L0 = b1.momentumL(Rcm) + b2.momentumL(Rcm)

        Aplicamos momentos opostos respeitando a terceira lei de Newton. Note
        que no problema com rotações devemos tomar cuidado em aplicar o impulso
        no mesmo ponto nos dois objetos. Caso isto não ocorra, haverá a
        produção de um torque externo, violando a lei de conservação do momento
        angular.

        >>> b1.apply_impulse(J, pos=(4, 0))
        >>> b2.apply_impulse(-J, pos=(4, 0))

        E verificamos que os valores finais são iguais aos iniciais

        >>> b1.momentumP() + b2.momentumP() == P0
        True
        >>> b1.momentumL(Rcm) + b2.momentumL(Rcm) == L0
        True
        '''

        if y is None:
            impulse = asvector(impulse_or_x)
        else:
            impulse = Vec2(impulse_or_x, y)
        self.boost(impulse * self._invmass)

        if pos is not None and self._invinertia:
            R = asvector(pos)
            R = R if relative else R - self.pos
            self.aboost(R.cross(impulse) * self._invinertia)

    # Variáveis angulares #####################################################
    def rotate(self, theta):
        '''Rotaciona o objeto por um ângulo _theta'''

        self._theta += theta
        if theta != 0.0:
            self.flags |= flags.dirty_any

    def aboost(self, delta):
        '''Adiciona um valor delta à velocidade angular'''

        self._omega += delta

    def vpoint(self, pos_or_x, y=None, relative=False):
        '''Retorna a velocidade linear de um ponto preso rigidamente ao
        objeto.

        Se o parâmetro `relative` for verdadeiro, o vetor de entrada é
        interpretado como a posição relativa ao centro de massa. O padrão é
        considerá-lo como a posição absoluta no sistema de coordenadas do
        mundo.

        Example
        -------

        Criamos um objeto inicialmente sem rotação

        >>> b1 = Body(pos=(1, 1), vel=(1, 1), mass=1, inertia=1)

        Neste caso, a velocidade relativa a qualquer ponto corresponde à
        velocidade do centro de massa

        >>> b1.vpoint(0, 0), b1.vpoint(1, 1)
        (Vec(1, 1), Vec(1, 1))

        Quando aplicamos uma rotação, a velocidade em cada ponto assume uma
        componente rotacional.

        >>> b1.aboost(1.0)
        >>> b1.vpoint(0, 0), b1.vpoint(1, 1)
        (Vec(2, 0), Vec(1, 1))

        Note que quando o parâmetro ``relative=True``, as posições são
        calculadas relativas ao centro de massa do objeto

        >>> b1.vpoint(0, 0, relative=True), b1.vpoint(1, 1, relative=True)
        (Vec(1, 1), Vec(0, 2))
        '''

        if y is None:
            pos = asvector(pos_or_x)
        else:
            pos = Vec2(pos_or_x, y)

        if not relative:
            pos -= self.pos

        return self._vel + pos.perp() * self._omega

    def orientation(self, theta=0.0):
        '''Retorna um vetor unitário na direção em que o objeto está orientado.
        Pode aplicar um ângulo adicional a este vetor fornecendo o parâmetro
        _theta.'''

        theta += self._theta
        return Vec2(cos(theta), sin(theta))

    def torque(self, t):
        '''Define uma torque externo análogo ao método .force()'''

        return 0.0

    def init_alpha(self):
        '''Inicializa o vetor de aceleração angular com os valores devido ao
        amortecimento'''

        self._alpha = - self._adamping * self._omega

    def apply_torque(self, torque, dt):
        '''Aplica um torque durante um intervalo de tempo dt.'''

        self.apply_alpha(torque * self._invinertia, dt)

    def apply_alpha(self, alpha, dt):
        '''Aplica uma aceleração angular durante um intervalo de tempo dt.

        Tem efeito em objetos cinemáticos.'''

        if alpha is None:
            alpha = self._alpha
        self.aboost(alpha * dt)
        self.rotate(self._omega * dt + alpha * dt ** 2 / 2.)

    def apply_aimpulse(self, angular_delta):
        '''Aplica um impulso angular ao objeto.

        O parâmetro de entrada corresponde à variação do momento angular do
        objeto.
        '''

        self.aboost(angular_delta * self._invinertia)


def vec_property(slot):
    '''Fabrica um slot que força a conversão de uma variável para a classe
    vetor.'''

    getter = slot.__get__
    setter = slot.__set__

    class VecProperty(object):
        __slots__ = []

        def __set__(self, obj, value):
            if not isinstance(value, Vec2):
                value = asvector(value)
            setter(obj, value)

        def __get__(self, obj, cls):
            if obj is None:
                return self
            else:
                return getter(obj, cls)

    return VecProperty()

Body.pos = vec_property(Body._pos)
Body.vel = vec_property(Body._vel)
Body.accel = vec_property(Body._accel)

###############################################################################
# Linear rigid bodies
###############################################################################
# TODO: remover esta classe e se basear no comportamento da flag can_rotate


class LinearRigidBody(Body):

    '''
    Classe que implementa corpos rígidos sem dinâmica angular.
    '''

    __slots__ = []
    DEFAULT_FLAGS = Body.DEFAULT_FLAGS & (flags.full ^ flags.can_rotate)

    def __init__(self, pos=(0, 0), vel=(0, 0),
                 mass=None, density=None, **kwds):

        super(LinearRigidBody, self).__init__(pos, vel, 0.0, 0.0,
                                              mass=mass, density=density,
                                              inertia='inf',
                                              flags=self.DEFAULT_FLAGS, **kwds)

    @property
    def inertia(self):
        return INF

    @inertia.setter
    def inertia(self, value):
        if float(value) != INF:
            raise ValueError('LinearObjects have infinite inertia, '
                             'got %r' % value)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



'''
Algumas funções úteis relacionadas com a física.
'''
from __future__ import division
from FGAme.mathtools import null2D


def aslist(args):
    N = len(args)
    if N == 1:
        return args[0]
    elif N == 0:
        raise TypeError('must give at least one argument')
    else:
        return args


def center_of_mass(*args):
    '''Calcula o centro de massa dos argumentos.

    Cada argumento deve ser um objeto com os parâmetros `mass` e `pos`

    Example
    -------

    >>> from FGAme.physics import Body
    >>> b1 = Body(pos=(0, 30), mass=1)
    >>> b2 = Body(pos=(10, 0), mass=3)
    >>> center_of_mass(b1, b2)
    Vec(7.5, 7.5)
    '''

    Rcm = null2D
    M = 0

    for obj in aslist(args):
        Rcm += obj.pos * obj.mass
        M += obj.mass

    return Rcm / M


def mass(*args):
    '''Calcula a soma das massas dos argumentos'''

    return sum(obj.mass for obj in aslist(args))


def inertia(*args, **_3to2kwargs):
    if 'pos' in _3to2kwargs: pos = _3to2kwargs['pos']; del _3to2kwargs['pos']
    else: pos = None
    '''Calcula o momento de inércia total dos argumentos.

    O argumento opcional `pos` descreve o ponto de referência para o cálculo
    do centro de inércia. O comportamento padrão é utilizar o centro de massa.

    Example
    -------

    >>> from FGAme.physics import Body
    >>> b1 = Body(pos=(0, 2), mass=1, inertia=2)
    >>> b2 = Body(pos=(1, 0), mass=3, inertia=1)
    >>> inertia(b1, b2)
    6.75
    '''

    args = aslist(args)
    if pos is None:
        pos = center_of_mass(args)

    inertia = 0
    for obj in args:
        inertia += obj.mass * (obj.pos - pos).norm_sqr() + obj.inertia
    return inertia


def momentumP(*args):
    '''Calcula o momentum linear total de todos os objetos fornecidos.

    O momentum linear total de um grupo de objetos consiste na soma do momentum
    de cada objeto.'''

    return sum((obj.momentumP() for obj in aslist(args)), null2D)


def momentumL(*args, **_3to2kwargs):
    if 'pos' in _3to2kwargs: pos = _3to2kwargs['pos']; del _3to2kwargs['pos']
    else: pos = None
    '''Calcula o momentum angular total de todos os objetos dados.

    O argumento opcional `pos` descreve o ponto de referência para o cálculo
    do momentum linear. O comportamento padrão é utilizar o centro de massa.

    O momentum angular total de um grupo de objetos consiste na soma do
    momentum de cada objeto.'''

    args = aslist(args)
    if pos is None:
        pos = center_of_mass(args)
    return sum((obj.momentumL(pos) for obj in args), null2D)


def energyK(*args):
    '''Calcula a energia cinética total dos objetos dados. Inclui a
    componente linear e a componente angular da energia.'''

    return sum(obj.energyK() for obj in aslist(args))


def linearK(*args):
    '''Calcula a componente linear da energia cinética dos objetos dados.

    A componente linear corresponde aos termos do tipo $m v^2 / 2$.'''

    return sum(obj.energyK() for obj in aslist(args))


def angularK(*args):
    '''Calcula a componente angular da energia cinética dos objetos dados.

    A componente linear corresponde aos termos do tipo $I \omega^2 / 2$,
    onde I representa o momento de inércia com relação ao centro de massa.'''

    return sum(obj.energyK() for obj in aslist(args))


def energyU(*args):
    '''Calcula a energia potencial total dos objetos dados.'''

    # TODO: incluir energia de interação entre as partículas
    return sum(obj.energyU() for obj in aslist(args))


def energy(*args):
    '''Calcula a energia total dos objetos dados.'''

    return energyK(*args) + energyU(*args)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import weakref
from collections import namedtuple
from FGAme.events.util import pyname, wrap_handler

ControlElem = namedtuple('ControlElem', ['idx', 'owner', 'wrapped', 'handler',
                                         'args', 'kwargs'])


class Signal(object):

    '''Representa um sinal comum anunciado por um objeto.'''

    sig_args = ()

    def __init__(self, name, num_args=0):
        self.name = name
        self.num_args = num_args
        self._num_filters = 0
        self._ctl_name = '_%s_ctl' % pyname(self.name)

    def __get__(self, instance, cls=None):
        if instance is not None:
            return getattr(instance, self._ctl_name)
        else:
            return self

    def __set__(self, instance, value):
        raise AttributeError('not writable')

    def __repr__(self):
        tname = type(self).__name__
        return '%s(%r, %s)' % (tname, self.name, self.num_args)

    def default_control(self, owner):
        '''Retorna o objeto tipo SignalCtl apropriado para essa classe de
        sinais'''

        return SignalCtl(self, owner)

    ###########################################################################
    # Fábrica de funções para os métodos listen_<sinal> e trigger_<sinal>
    ###########################################################################

    def _factory_listen_method(self):
        signal = self  # não confundir com o self da função fabricada!

        def listen_method(self, *args, **kwargs):
            '''vsds %(signal)s'''

            # Confere argumentos de entrada
            args_ = kwargs.pop('args', None)
            kwargs_ = kwargs.pop('kwargs', None)
            owner = kwargs.pop('owner', None)
            if args_:
                args = args_
            if kwargs_:
                kwargs = kwargs_

            # Extrai os filtros, caso necessário
            filter_args = args[:signal._num_filters]
            try:
                handler = args[signal._num_filters]
            except IndexError:
                handler = None
            args = args[signal._num_filters + 1:]

            if args_ is not None and args:
                raise TypeError('cannot specify the args parameter and also '
                                'insert positional arguments')
            if kwargs_ is not None and kwargs:
                raise TypeError('cannot specify the kwargs parameter and also '
                                'insert keyword arguments')

            # Caso handler=None, estamos chamando o método listen na versão
            # de decorador
            if handler is None:
                def decorator(func):
                    args_ = filter_args + (func,)
                    kwargs_ = {'args': args, 'kwargs': kwargs, 'owner': owner}
                    return listen_method(self, *args_, **kwargs_)
                return decorator

            # Registra nas listas de controle
            wrapped = wrap_handler(
                handler, args, kwargs, pre_args=signal.num_args)
            ctl = getattr(self, '_%s_ctl' % pyname(signal.name))

            if signal._num_filters == 0:
                idx = len(ctl._data)
                elem = ControlElem(idx, owner, wrapped, handler, args, kwargs)
                ctl._runner.append(wrapped)
                ctl._data.append(elem)
                return idx

            elif signal._num_filters == 1:
                key = filter_args[0]
                ids = {}

                if key is None:
                    # Registra como handler de todas os filtros no livro
                    for k in ctl._runner:
                        if k is None:
                            continue
                        wrapped_i = wrap_handler(handler, (k,) + args, kwargs,
                                                 pre_args=signal.num_args)
                        ids[k] = listen_method(self, k, wrapped_i)

                    # Cria um elemento de controle substituindo o campo
                    # "wrapped" pelo dicionário das ids em cada filtro
                    control = ctl._data.setdefault(None, [])
                    book = ctl._runner.setdefault(None, [])
                    idx = len(control)
                    elem = ControlElem(idx, owner, ids, handler, args, kwargs)
                    control.append(elem)
                    book.append(
                        wrap_handler(
                            handler, args, kwargs, pre_args=signal.num_args))

                    return (None, idx)
                else:
                    try:
                        idx = len(ctl._data[key])
                    except KeyError:
                        ctl._data[key] = []
                        ctl._runner[key] = []

                        for ctl_e in ctl._data.get(None, []):
                            wrapped_i = wrap_handler(
                                ctl_e.handler, (key,) + ctl_e.args,
                                ctl_e.kwargs, signal.num_args)
                            ctl_e.wrapped[key] = listen_method(
                                self, key, wrapped_i)

                        idx = len(ctl._data[key])

                    elem = ControlElem(
                        idx, owner, wrapped, handler, args, kwargs)
                    ctl._runner[key].append(wrapped)
                    ctl._data[key].append(elem)
                    return key, idx

            else:
                book = ctl._runner[filter_args]
                control = ctl._data[filter_args]
            raise RuntimeError

        listen_method.__name__ = 'listen_' + pyname(signal.name)
        listen_method.__doc__ = listen_method.__doc__ % {'signal': signal.name}
        return listen_method

    def _factory_trigger_method(self):
        signal = self  # não confundir com o self da função fabricada!
        attr_name = pyname('_%s_book' % signal.name)

        # Especializamos para 0, 1 ou 2 argumentos por questões de performance
        if signal.num_args == 0:
            def trigger_method(self):
                book = getattr(self, attr_name)
                for wrapped_func in book:
                    wrapped_func()

        elif signal.num_args == 1:
            def trigger_method(self, arg):
                book = getattr(self, attr_name)
                for wrapped_func in book:
                    wrapped_func(arg)

        elif signal.num_args == 2:
            def trigger_method(self, arg1, arg2):
                book = getattr(self, attr_name)
                for wrapped_func in book:
                    wrapped_func(arg1, arg2)

        else:
            def trigger_method(self, *args):
                if len(args) != signal.num_args:
                    raise TypeError('expected %s arguments' % self.num_args)

                book = getattr(self, attr_name)
                for wrapped_func in book:
                    wrapped_func(*args)

        trigger_method.__name__ = 'trigger_' + pyname(signal.name)
        # trigger_method.__doc__ = trigger_method.__doc__ % \
        #                         {'signal': signal.name}
        return trigger_method


class FilteredSignal(Signal):

    '''Implementa um sinal filtrado: ou seja, um sinal que assume um segundo
    argumento e dispacha apenas os handlers registrados para aquele argumento.

    Útil para eventos do tipo key-press onde podemos acionar os handlers
    seletivamente de acordo com a tecla pressionada.
    '''

    def __init__(self, name, num_args=0, filter_names=None):
        super(FilteredSignal, self).__init__(name, num_args)
        self._num_filters = 1
        self._filter_names = filter_names or (None,) * self._num_filters

    def default_control(self, owner):
        return SignalCtlMulti(self, owner)

    def _factory_trigger_method(self):
        signal = self  # não confundir com o self da função fabricada!
        attr_name = pyname('_%s_book' % signal.name)

        # Especializamos para 0, 1 ou 2 argumentos por questões de performance
        if signal.num_args == 0:
            def trigger_method(self, key):
                book = getattr(self, attr_name)
                kbook = book.get(key, None)
                if kbook is not None:
                    for wrapped_func in kbook:
                        wrapped_func()
                else:
                    gen_book = book.get(None, None)
                    if gen_book is not None:
                        for func in gen_book:
                            func(key)

        elif signal.num_args == 1:
            def trigger_method(self, key, arg):
                book = getattr(self, attr_name)
                kbook = book.get(key, None)
                if kbook is not None:
                    for wrapped_func in kbook:
                        wrapped_func(arg)
                else:
                    gen_book = book.get(None, None)
                    if gen_book is not None:
                        for func in gen_book:
                            func(key, arg)

        else:
            def trigger_method(self, key, *args):
                if len(args) != signal.num_args:
                    raise TypeError('expected %s arguments' % self.num_args)

                book = getattr(self, attr_name)
                kbook = book.get(key, None)
                if kbook is not None:
                    for wrapped_func in kbook:
                        wrapped_func(*args)
                else:
                    gen_book = book.get(None, None)
                    if gen_book is not None:
                        for func in gen_book:
                            func(key, *args)

        trigger_method.__name__ = 'trigger_' + pyname(signal.name)
        # trigger_method.__doc__ = (trigger_method.__doc__
        #                            % {'signal': signal.name})
        return trigger_method

    def __repr__(self):
        names = ''
        if self._filter_names:
            names = ', filter_names=%s' % list(self._filter_names)
        tname = type(self).__name__
        return '%s(%r, %s%s)' % (tname, self.name, self.num_args, names)


class DelegateSignal(Signal):

    '''Delega o sinal para um atributo da classe.

    Sinais deste tipo não podem ser emitidos diretamente, mas é possível
    registrar handlers que serão encaminhados para o método listen do atributo
    responsável por gerenciar o despacho de eventos'''

    def __init__(self, name, delegate_to, num_filters=0, filter_names=None):
        super(DelegateSignal, self).__init__(name)
        self.delegate_to = delegate_to
        self._num_filters = num_filters
        self._filter_names = filter_names

    def _factory_listen_method(self):
        signal = self

        def listen_method(self, *args, **kwds):
            kwds.setdefault('owner', self)
            delegate_to = getattr(self, signal.delegate_to)
            return delegate_to.listen(signal.name, *args, **kwds)

        return listen_method

    def _factory_trigger_method(self):

        def trigger_method(self, *args):
            raise RuntimeError('signals can only be trigged by the owner')

        return trigger_method

    def __get__(self, instance, cls=None):
        if instance is not None:
            owner = getattr(instance, self.delegate_to)
            return getattr(owner, self._ctl_name)
        else:
            return self


###############################################################################
#                           Controle de sinais
###############################################################################
class SignalCtl(object):

    '''A classe SignalCtl implementa uma interface para gerenciar os handlers
    registrados para um determinado sinal'''

    def __init__(self, signal, instance):
        self.signal = signal
        self._data = []
        self._runner = []
        #self._instance = weakref.ref(instance)
        self._instance = instance

    @property
    def instance(self):
        # return self._instance()
        return self._instance

    @property
    def name(self):
        return self.signal.name

    @property
    def handlers(self):
        return [ctl.handler for ctl in self._data if ctl is not None]

    @property
    def handlers_full(self):
        return [(ctl.handler, ctl.args, ctl.kwargs)
                for ctl in self._data if ctl is not None]

    @property
    def owners(self):
        return [ctl.owner for ctl in self._data if ctl is not None]

    def remove(self, *args, **kwds):
        '''Remove handlers baseado em algum critério.

        Assinaturas
        -----------

        obj.remove(id=<handler_id>) ou obj.remove(<handler_id>)
            Remove handler especificando a id retornada pelo método listen()

        obj.remove(handler=<handler function>)
            Remove handler a partir da função handler registrada no método
            listen(). Esta opção remove todos os handlers associados à função
            dada, mesmo que registrados com argumentos diferentes

        obj.remove(owner=<owner>)
            Remove todos os handlers associados ao dono fornecido. Esta função
            é útil em delegações em que os handlers podem ser criados por
            vários objetos diferentes.
        '''

        if not args:
            key, value = kwds.popitem()
        elif len(args) == 1:
            key, value = 'id', args[0]
        else:
            raise TypeError('wrong number of positional arguments')
        if kwds:
            raise TypeError('must specify exactly one criteria')

        removed = []

        if key == 'id':
            idx = value
            wrapped = self._data[idx].wrapped
            w_idx = [i for i, w in enumerate(self._runner) if w is wrapped]
            if len(w_idx) > 1:
                raise NotImplementedError

            del self._runner[w_idx[0]]
            removed.append(self._data[idx])
            self._data[idx] = None

        elif key == 'handler':
            for i, ctl in enumerate(self._data):
                if ctl.handler == value:
                    removed_obj = self.remove(id=i)
                    removed.extend(removed_obj)

        elif key == 'owner':
            for i, ctl in enumerate(self._data):
                if ctl.owner == value:
                    removed_obj = self.remove(id=i)
                    removed.extend(removed_obj)

        else:
            raise TypeError('invalid argument: %s' % key)

        return removed

    def restore(self, L):
        '''Restaura uma lista de sinais removidas pelo método remove()'''

        for ctl_elem in L:
            if self._data[ctl_elem.idx] is not None:
                raise RuntimeError(
                    'restoring handler that were not removed or '
                    'were already restored')
            else:
                self._data[ctl_elem.idx] = ctl_elem
                self._runner.append(ctl_elem.wrapped)

    def listen(self, *args, **kwds):
        '''Registra um callback para o sinal'''

        return self.instance.listen(self.name, *args, **kwds)

    def trigger(self, *args, **kwds):
        '''Emite um sinal'''

        return self.instance.trigger(self.name, *args, **kwds)

    #
    # Magic methods
    #
    def __contains__(self, handler):
        return handler in self._data


class SignalCtlMulti(SignalCtl):

    '''Gerencia handlers de sinais filtrados'''

    # TODO: talvez implementar como um dicionario de SignalCtl

    def __init__(self, signal, owner):
        super(SignalCtlMulti, self).__init__(signal, owner)
        self._runner = {}
        self._data = {}



from __future__ import division
from FGAme import schedule_iter_dt, schedule, unschedule
from FGAme.mathtools import Vec2, cos, sin
from FGAme.draw import Color


class RunnerControl(object):

    def __init__(self, action):
        self.is_cancelled = False
        self.action = action

    def _assure_cancelled(self):
        if self.is_cancelled:
            raise RuntimeError('cannot operate in a cancelled action')

    def cancel(self):
        unschedule(self.action)
        self.is_cancelled = True

    def pause(self):
        self._assure_cancelled()
        unschedule(self.action)

    def resume(self):
        schedule(self.action)


class Action(object):

    '''Classe mãe de todas as ações.


    Usamos a convenção que classes abstratas são nomeadas em CamelCase enquanto
    classes que realizam ações concretas são nomeadas como funções.
    '''
    instant = False

    def copy(self):
        '''Retorna copia de si mesmo'''

    def prepare(self, target=None, **kwds):
        '''Prepara a ação para iniciar a execução.'''

    def finish(self, target, **kwds):
        ''''''
        pass

    def start(self, target=None, **kwds):
        '''Executa a ação'''

        return schedule_iter_dt(self.runner, target, **kwds)

    def runner(self, dt, target=None, **kwds):
        raise NotImplementedError

    #
    # Define operações
    #
    def __rshift__(self, action):
        return sequence(self, action)

    def __mul__(self, other):
        if not isinstance(other, int):
            raise TypeError("Can only multiply actions by ints")
        if other == 0:
            return DoNothingAction()
        elif other == 1:
            return self
        elif other > 1:
            return loop(self, other)
        else:
            raise ValueError('cannot loop a negative number of times')

    def __or__(self, action):
        return parallel(self, action)

    def __reversed__(self):
        raise Exception("Action %s cannot be reversed" % type(self).__name__)


class DoNothingAction(Action):

    def runner(self, *args, **kwds):
        yield


class ActionModifier(Action):

    def __init__(self, action):
        if not isinstance(action, Action):
            raise TypeError('must be initialized with an Action object')
        self.action = action


class InstantAction(Action):
    instant = True

    def runner(self, dt, target=None, **kwds):
        self.execute(target, **kwds)
        yield


class IntervalAction(Action):
    instant = False

    def __init__(self, duration):
        self.duration = float(duration)


class delay(IntervalAction):

    def runner(self, dt, target=None, **kwds):
        delay = self.duration
        time = 0.0
        while time < delay:
            time += dt
            yield


class sequence(Action):

    def __init__(self, *actions):
        self.actions = list(actions)

    def runner(self, dt, target=None, **kwds):
        for action in self.actions:
            if action.instant:
                action.execute(target, **kwds)
            else:
                for _ in action.runner(dt, target, **kwds):
                    yield


class parallel(Action):

    def __init__(self, *actions):
        self.actions = list(actions)

    def runner(self, dt, target=None, **kwds):
        iterators = [action.runner(dt, target, **kwds)
                     for action in self.actions]
        while iterators:
            finished = []
            for (idx, runner) in enumerate(iterators):
                try:
                    next(runner)
                except StopIteration:
                    finished.append(idx)
            if finished:
                for idx in reversed(finished):
                    iterators.pop(idx)


class move_to(IntervalAction):

    def __init__(self, pos, interval):
        super(move_to, self).__init__(interval)
        self.pos = Vec2(*pos)

    def runner(self, dt, target, **kwds):
        final_pos = self.pos
        interval = self.duration
        time = 0

        while time < interval:
            time += dt
            try:
                fraction = min(dt / (interval - time), 1.0)
            except ZeroDivisionError:
                fraction = 1.0

            if fraction <= 0.0:
                return
            else:
                target.pos += (final_pos - target.pos) * fraction
                yield


class move_by(move_to):

    def runner(self, dt, target, **kwds):
        time = 0
        interval = self.duration
        delta = self.pos * (dt / interval)

        while time < interval:
            time += dt
            target.pos += delta
            yield


class place_at(InstantAction):

    def __init__(self, x, y=None):
        if y is None:
            self.pos = Vec2(*x)
        else:
            self.pos = Vec2(x, y)

    def execute(self, target, **kwds):
        target.pos = self.pos


class hide(InstantAction):

    def execute(self, target, **kwds):
        target.hide()


class show(InstantAction):

    def execute(self, target, **kwds):
        target.show()


class toggle_visibility(InstantAction):

    def execute(self, target, **kwds):
        if target.visible:
            target.hide()
        else:
            target.show()


class stop(InstantAction):

    def execute(self, target, **kwds):
        target.vel *= 0


class call_function(InstantAction):

    def __init__(self, function, args=(), kwds={}, pass_target=False):
        self.function = function
        self.args = tuple(args)
        self.kwds = dict(kwds)
        self.pass_target = pass_target

    def execute(self, target=None, **kwds):
        args = self.args
        args = (target,) + args if self.pass_target else args
        kwds = self.kwds
        self.function(*args, **kwds)


class loop(ActionModifier):

    def __init__(self, action, iterations=None):
        super(loop, self).__init__(action)
        self.iterations = float('inf') if iterations is None else iterations

    def runner(self, *args, **kwds):
        i = 0
        while i < self.iterations:
            i += 1
            for _ in self.action.runner(*args, **kwds):
                yield


class ping_pong(ActionModifier):

    def runner(self, *args, **kwds):
        for _ in self.action.runner(*args, **kwds):
            yield
        for _ in reversed(self.action).runner(*args, **kwds):
            yield


class evolve(Action):

    def __init__(self, property_name, evolution, interval, relative=True):
        self.property_name = property_name
        self.evolution = evolution
        self.interval = interval
        self.relative = relative

    def runner(self, dt, target, **kwds):
        property_name = self.property_name
        evolution = self.evolution
        interval = self.interval
        relative = self.relative
        starting_value = getattr(target, property_name)
        time = 0

        while time < interval:
            time += dt
            new_value = evolution(time)
            if relative:
                new_value -= starting_value
            setattr(target, property_name, new_value)


class set_trajectory(Action):

    def __init__(self, trajectory, interval, relative=True):
        self.trajectory = trajectory
        self.interval = interval
        self.relative = relative

    def runner(self, dt, target, **kwds):
        time = 0
        interval = self.interval
        position = self.trajectory
        pos_shift = Vec2(0, 0) if not self.relative else target.pos

        while time < interval:
            time += dt
            target.pos = pos_shift + position(time)


class set_attribute(InstantAction):

    def __init__(self, attribute, value):
        self.attribute = attribute
        self.value = value

    def execute(self, target, **kwds):
        setattr(target, self.attribute, self.value)


class set_color(InstantAction):

    def __init__(self, color):
        self.color = Color(color)

    def execute(self, target, **kwds):
        target.color = self.color


class set_velocity(set_attribute):

    def __init__(self, velocity, vy=None):
        if vy is not None:
            velocity = (velocity, vy)
        super(set_velocity, self).__init__('vel', velocity)


class color_effect(IntervalAction):

    def __init__(self, effect, interval, *args, **kwds):
        super(color_effect, self).__init__(interval)
        self.effect = effect
        self.args = tuple(args)
        self.kwds = dict(kwds)

    def runner(self, dt, target, **kwds):
        target.color = self.color


class time_map(ActionModifier):

    def __init__(self, action, mapping=None):
        super(time_map, self).__init__(action)
        self.mapping = mapping or self._default_mapping

    def _default_mapping(self, tau):
        return tau


class accelerate(time_map):

    def _default_mapping(self, tau):
        return tau ** 2


class deaccelerate(time_map):

    def _default_mapping(self, tau):
        return 1 - (tau - 1) ** 2


class finish_game(InstantAction):

    def __init__(self, *args):
        self.args = args

    def execute(self, *args, **kwds):
        raise SystemExit(*args)


class destroy(InstantAction):

    def execute(self, target, **kwds):
        target.destroy()


class add_to(InstantAction):

    def __init__(self, world):
        self.world = world

    def execute(self, target, **kwds):
        self.world.add(target)


class from_generator(Action):

    def __new__(self, generator=None, *args, **kwds):
        if generator is None or generator is Ellipsis:
            def decorator(generator):
                return from_generator(generator, *args, **kwds)
            return decorator

        new = object.__new__(from_generator)
        new.generator = generator
        new.args = tuple(args)
        return new
        self.kwds = dict(kwds)

    def runner(self, *args, **kwds):
        for _ in self.generator(*args, **kwds):
            yield


class step_forward(IntervalAction):

    def __init__(self, step, duration):
        super(step_forward, self).__init__(duration)
        self.step = step

    def runner(self, dt, target, **kwds):
        n_steps = dt / self.duration
        delta = self.step / n_steps

        for _ in range(n_steps):
            target.pos += Vec2.from_polar(delta, target.theta)


if __name__ == '__main__':
    from FGAme import *
    w = World()
    circles = [draw.Circle(10, pos=pos.middle) for _ in range(20)]
    w.add(circles)

    action = (delay(0.5) >> place_at(0, 0) >> delay(0.5)
              >> move_to(pos.middle, 1) >> set_color('red'))

    for idx, c in enumerate(circles):
        (delay(0.1 + 0.1 * idx) >> action).start(c)

    w.run()



# -*- coding: utf-8 -*-
from __future__ import division
import time
import functools
from collections import deque, namedtuple
from FGAme.events import EventDispatcher, signal


TimedAction = namedtuple('TimedAction', ['time', 'action', 'args', 'kwds'])


class MainLoop(EventDispatcher):

    '''Implementa o loop principal de jogo.

    Tarefas do Loop principal:
        * Realizar a leitura de entradas do usuário (mouse, teclado, etc) e
          executar ações pertinentes.
        * Atualizar o estado de jogo
        * Atualizar a visualização
        * Tentar manter uma taxa de atualização fixa (se a capacidade de
          processamento do computador permitir)
        * Agendar tarefas
        * Gerenciar os estados de jogo (ex., coordenar entre estados de menu,
          jogo, configuração, etc)

    Parameters
    ----------

    screen : Screen
        Objeto do tipo Screen já inicializado.
    input : Screen
        Objeto do tipo Input já inicializado.
    fps : float
        Número de quadros por segundo que a simulação de física deve
        tentar rodar.
    time : float (padrão=0.0)
        Tempo de simulação a partir do do método run(). Pode defasar com
        relação ao tempo real devido à eventos de "frame-skip"
    n_iter : int (padrão=0)
        Número de frames processados. É incrementado a cada passo de
        simulação.
    clock : Clock
        Objeto do tipo clock. Serve como controle para a execução de ações
        programadas.

    Attributes
    ----------

    Todos os parâmetros estão disponíveis como atributos. Além destes, temos

    dt : float
        Recíproco de fps, mostra o tempo esperado gasto em cada frame.
    n_skip : int
        Número de frames pulados durante a simulação.
    sleep_time : float
        Tempo necessário dormir que manteria a taxa de atualização constante.


    Sinais
    ------

    pre-draw
        Emitido antes de iniciar a renderização principal, mas após pintar o
        fundo. O callback possui a assinatura callback(screen), onde `screen`
        é o objecto que realiza a renderização na tela no backend selecionado.
    post-draw
        Semelhante ao pre-draw, mas executado após a renderização principal.
    frame-skip
        Acionado sempre que houver uma perda de frames na simulação. O callback
        é chamado como callback(dt), onde dt é o tempo que extrapolou o máximo
        para aquele frame, em segundos. Ex.: se cada frame deve rodar em 0.1s e
        um determinado frame gastou 0.15s, então o callback é acionado com 0.05
        no argumentoframe

    '''

    pre_draw = signal('pre-draw', num_args=1)
    post_draw = signal('post-draw', num_args=1)
    frame_skip = signal('frame-skip', num_args=1)
    frame_enter = signal('frame-enter')
    frame_frame_leave = signal('frame-leave')
    _instance = None

    def __init__(self, screen, input, fps=None):  # @ReservedAssignment
        super(MainLoop, self).__init__()
        self.screen = screen
        self.input = input
        self.fps = fps or 60
        self.dt = 1.0 / self.fps
        self.time = 0.0
        self.sleep_time = 0.0
        self.n_iter = 0
        self.n_skip = 0
        self._running = False
        self._action_queue = deque()

        if self._instance is not None:
            raise RuntimeError('MainLoop is a singleton')
        else:
            MainLoop._instance = self

    def run(self, state, timeout=None, maxiter=None, wait=True):
        '''Roda o loop principal.

        Parameters
        ----------

        state :
            Objeto que controla o estado principal do jogo ou aplicação.
            Deve suportar o método state.update(dt), que atualiza o estado
            dado.
        timeout : float
            Executa a simulação até o tempo máximo especificado.
        maxiter : int
            Define um número máximo de iterações.
        wait : bool
            Se wait=False, renderiza e atualiza cada frame imediatamente após
            o outro. Isto desabilita o controle de frame-rate e é utilizado
            apenas para depuração.
        '''

        timeout = float('inf') if timeout is None else float(timeout)
        timeout += self.time
        maxiter = float('inf') if maxiter is None else maxiter
        maxiter += self.n_iter

        self._running = True
        while self._running:
            self.step(state, wait=wait)

            # Verifica que já ultrapassou o tempo de simulação
            if timeout is not None and self.time >= timeout:
                break
            if maxiter is not None and self.n_iter >= maxiter:
                break

    def step(self, state, wait=True):
        '''Realiza um passo de simulação sobre o estado fornecido'''

        start_time = time.time()
        self.trigger_frame_enter()

        # Executa ações one_shot no início dos frames
        Q = self._action_queue
        while Q and (self.time > Q[0].time):
            x = Q.popleft()
            x.action(*x.args, **x.kwds)

        # Captura entrada do usuário
        self.input.poll()  # @UndefinedVariable

        # Atualiza o estado (física e animações) de acordo
        state.update(self.dt)

        # Desenha os objetos na tela
        screen = self.screen
        bg_color = getattr(state, 'background', 'white')
        if bg_color is not None:
            screen.clear_background(bg_color)
        self.trigger_pre_draw(screen)
        state.get_render_tree().draw(screen)
        self.trigger_post_draw(screen)
        screen.flip()

        # Finaliza o frame
        time_interval = time.time() - start_time
        self.sleep_time = self.dt - time_interval
        self.trigger_frame_leave()

        # Recalcula o tempo de dormir pois frame_leave() pode ter consumido
        # alguns ciclos
        time_interval = time.time() - start_time
        self.sleep_time = self.dt - time_interval
        if wait:
            if self.sleep_time > 0:
                time.sleep(self.sleep_time)
            else:
                self.n_skip += 1
                self.trigger_frame_skip(-wait)

        self.time += self.dt
        self.n_iter += 1

    def stop(self):
        '''Finaliza o loop principal'''

        self._running = False

    def one_shot(self, time, function=None, *args, **kwds):
        '''Agenda a execução da ação dada para acontecer imediatamente após
        transcorridos o número de segundos fornecidos. A ação será executada
        logo após o fim de cada passo, antes de iniciar o frame seguinte.
        '''

        # Decorator form
        if function is None or function is Ellipsis:
            def decorator(func):
                return self.one_shot(time, func, *args, **kwds)
            return decorator

        Q = self._action_queue
        Q.append(TimedAction(self.time + time, function, args, kwds))
        self._action_queue = deque(sorted(Q, key=lambda x: x.time))

    def one_shot_frames(self, n_frames, function=None, *args, **kwds):
        '''Similar a one_shot, mas agenda a execução para ocorrer em um número
        de frames especificado
        '''

        return self.one_shot(n_frames * self.dt, function, *args, **kwds)

    def one_shot_absolute(self, time, function=None, *args, **kwds):
        '''Similar a one_shot, mas agenda a execução para ocorrer em um valor
        absoluto na linha de tempo e não após transcorrido o intervalo dado.
        '''

        return self.one_shot(time - self.time, function, *args, **kwds)

    def periodic(self, time_interval, function=None, *args, **kwds):
        '''Similar à one_shot, mas agenda a simulação para ocorrer
        repetidamente cada vez que passar o intervalo de tempo fornecido.'''

        # Decorador
        if function is None or function is Ellipsis:
            def decorator(func):
                return self.periodic(time_interval, func, *args, **kwds)
            return decorator

        def recursive_action(*r_args, **r_kwds):
            function(*r_args, **r_kwds)
            self.one_shot(time_interval, recursive_action, *args, **kwds)

        self.one_shot(0, recursive_action, *args, **kwds)

    def periodic_frames(self, n_frames, function=None, *args, **kwds):
        '''Similar à periodic(), mas agenda o intervalo de execução em frames
        ao invés de segundos'''

        self.periodic(n_frames * self.dt, function, *args, **kwds)

    def schedule(self, function=None, *args, **kwds):
        '''Agenda função para ser executada em cada frame.

        Parameters
        ----------

        function : callable
            Função a ser executada em cada frame. Por padrão, a função não
            recebe nenhum parâmetro, no entanto os parâmetros posicionais ou
            por nome adicionais são repassados.
        start : bool
            Determina se a ação deve ser executada no início ou ao final do
            frame. (padrão=True)
        skip : int ou tuple
            Determina quantos frames são pulados a cada execução. Por exemplo,
            se skip=2, a função é executada a cada dois frames. (padrão=1)
            Se for uma tupla (x, y), o primeiro valor representa o intervalo
            e o segundo representa quantos frames é necessário esperar para
            iniciar a execução.
'''

        # Decorator form
        if function is None or function is Ellipsis:
            def decorator(func):
                return self.schedule(func, *args, **kwds)
            return decorator

        skip = kwds.pop('skip', 1)
        if skip != 1:
            raise NotImplementedError

        if kwds.pop('start', True):
            return self.frame_enter.listen(function, *args, **kwds)
        else:
            return self.frame_leave.listen(function, *args, **kwds)

    def schedule_dt(self, function=None, *args, **kwds):
        '''Similar à `schedule()`, mas utiliza uma função que recebe o
        intervalo de tempo `dt` transcorrido entre cada frame como primeiro
        parâmetro.
        '''

        return self.schedule(function, self.dt, *args, **kwds)

    def schedule_optional(self, function, *args, **kwds):
        '''Agenda ação para executar ao final de cada frame se o mesmo não
        tiver estourado o tempo limite.

        Útil para executar ações de manutenção que podem ser adiadas para
        quando a CPU estiver mais livre.
        '''

        mainloop = MainLoop._instance

        def optional_action():
            if mainloop.sleep_tile:
                function()

        return self.schedule(optional_action, start=False)

    def schedule_optional_dt(self, function, *args, **kwds):
        '''Similar à `schedule_optional()`, mas utiliza uma função que recebe o
        intervalo de tempo `dt` transcorrido entre cada frame como primeiro
        parâmetro.
        '''

        self.schedule_optional(function, self.dt, *args, **kwds)

    def schedule_iter(self, iterator, *args, **kwds):
        '''Agenda a execução de um iterador ou gerador em um passo a cada frame.

        Quando o gerador for totalmente consumido, ele é eliminado do loop de
        execução.

        Parameters
        ----------

        iterator :
            Qualquer iterador ou gerador. O loop principal chamará
            repetidamente next(iterator) para cada frame executado.
        start : bool
            Determina se a ação será executada no início ou no fim do frame.
        skip : int
            Executa as ações a cada skip (padrão skip=1) frames.
        '''

        start = kwds.pop('start', True)
        skip = kwds.pop('skip', 1)

        # Inicia variáveis
        action_id = None
        idx = [1]

        # Aceita uma função geradora como entrada (funções com cláusula yield)
        try:
            iterator = iter(iterator)
        except TypeError:
            iterator = iterator(*args, **kwds)

        def step():
            try:
                if idx[0] % skip == 0:
                    next(iterator)
            except StopIteration:
                mainloop = MainLoop._instance
                if start:
                    mainloop.frame_enter.remove(action_id)
                else:
                    mainloop.frame_leave.remove(action_id)
            idx[0] += 1

        action_id = self.schedule(step, start=start)
        return action_id

    def schedule_iter_dt(self, iterator, *args, **kwds):
        '''Similar à `schedule_iter()`, mas utiliza uma função que recebe o
        intervalo de tempo `dt` transcorrido entre cada frame como primeiro
        parâmetro.
        '''

        self.schedule_iter(iterator, self.dt, *args, **kwds)

    def delayed(self, time, function=None, *args, **kwds):
        '''Semelhante à schedule, mas atrasa a execução da função pelo tempo
        especificado'''

        # Decorador
        if function is None:
            def decorator(func):
                return self.delayed(time, func, *args, **kwds)
            return decorator

        def do_schedule():
            self.schedule(function, *args, **kwds)

        return self.one_shot(time, do_schedule)

    def delayed_frames(self, n_frames, function=None, *args, **kwds):
        '''Semelhante à delayed(), mas especifica um número de frames ao invés
        do intervalo de tempo'''

        return self.delayed(self.dt * n_frames, function, *args, **kwds)

    def delayed_absolute(self, time, function=None, *args, **kwds):
        '''Semelhante à delayed(), mas inicia após a simulação atingir um valor
        absoluto de tempo.'''

        return self.delayed(time - self.time, function, *args, **kwds)

    def unschedule(self, handler):
        '''Remove uma ação da execução'''

        # TODO: make everything cancellable!
        mainloop = MainLoop._instance
        if mainloop is not None:
            mainloop.frame_enter.remove(handler)
            mainloop.frame_leave.remove(handler)
        else:
            raise RuntimeError('cannot unschedule action from simulation that '
                               'has not started')


#
# Global schedulers
#
def _make_scheduler(name):
    '''Cria uma função tipo "scheduler" que delega a ação para o objeto
    mainloop principal'''

    @functools.wraps(getattr(MainLoop, name))
    def scheduler_function(*args, **kwds):
        try:
            func = getattr(MainLoop._instance, name)
        except AttributeError:
            raise RuntimeError('MainLoop is not configured')
        return func(*args, **kwds)

    globals()[name] = scheduler_function
    return scheduler_function

one_shot = _make_scheduler('one_shot')
one_shot_frames = _make_scheduler('one_shot_frames')
one_shot_absolute = _make_scheduler('one_shot_absolute')
periodic = _make_scheduler('periodic')
periodic_frames = _make_scheduler('periodic_frames')
schedule = _make_scheduler('schedule')
schedule_optional = _make_scheduler('schedule_optional')
schedule_iter = _make_scheduler('schedule_iter')
schedule_dt = _make_scheduler('schedule_dt')
schedule_optional_dt = _make_scheduler('schedule_optional_dt')
schedule_iter_dt = _make_scheduler('schedule_iter_dt')
delayed = _make_scheduler('delayed')
delayed_frames = _make_scheduler('delayed_frames')
delayed_absolute = _make_scheduler('delayed_absolute')
unschedule = _make_scheduler('unschedule')


if __name__ == '__main__':
    from FGAme import World, Circle, pos
    w = World()
    c1 = Circle(10, world=w)
    c2 = Circle(10, color='red', world=w)
    c3 = Circle(10, color='blue', world=w)

    @schedule
    def move():
        c2.pos += (1, 1)

    @schedule_iter
    def move_circle():
        for _ in range(50):
            c1.pos += (2, 5)
            yield

    @periodic(1)
    def move_c3():
        c3.pos += (50, 30)

    @delayed(2)
    def move_circle2():
        c1.pos += (1, 0)

    @one_shot(3)
    def move_circle3():
        c1.pos = pos.middle

    w.run()



# -*- coding: utf8 -*-
'''
Este exemplo mostra um gás de esferas rígidas em contato com um êmbulo sujeito
a uma força viscosa. A energia é dissipada no movimento do êmbolo e aos poucos
as partículas cessam o movimento.

É possível adicionar uma força às partículas utilizando as setas do teclado.
'''

from __future__ import division
from FGAme import *
from random import uniform, randint


class Gas(World):

    def __init__(self,
                 gravity=200, friction=0.0, restitution=1.0,
                 num_balls=100, speed=200, radius=10,
                 color='random'):
        '''Cria uma simulação de um gás de wpartículas confinado por um êmbolo
        com `num_balls` esferas de raio `radius` com velocidades no intervalo
        de +/-`speed`.'''

        super(Gas, self).__init__(gravity=gravity, friction=friction)
        self.add_bounds(width=(10, 10, 10, -1000), delta=400)

        # Inicia bolas
        self.bolas = []
        for _ in range(num_balls):
            pos = Vec2(uniform(20, 780), uniform(20, 400))
            vel = Vec2(uniform(-speed, speed), uniform(-speed, speed))
            bola = Circle(radius=radius, vel=vel, pos=pos, mass=1,
                          color='random')
            self.bolas.append(bola)
            self.add(bola)

        # Inicia êmbolo
        embolo = AABB(bbox=(11, 789, 420, 470), color=(150, 0, 0),
                      mass=num_balls / 2, damping=5)
        self.add(embolo)

    @listen('long-press', 'q')
    def energy_up(self):
        '''Aumenta a energia de todas as partículas'''

        for bola in self.bolas:
            bola.vel += 1 * bola.vel.normalize()

    @listen('long-press', 'a')
    def energy_down(self):
        '''Diminui a energia de todas as partículas'''

        for bola in self.bolas:
            bola.vel *= 0.98

    @listen('key-down', 'space')
    def toggle_pause(self):
        super(Gas, self).toggle_pause()

    @listen('long-press', 'up', Vec2(0, 9))
    @listen('long-press', 'down', Vec2(0, -2))
    @listen('long-press', 'left', Vec2(-4, 0))
    @listen('long-press', 'right', Vec2(4, 0))
    def boost(self, Vec):
        for bola in self.bolas:
            bola.vel += Vec
            bola.vel *= 0.99

# Inicia a simulação
if __name__ == '__main__':
    game = Gas()
    game.run()



# -*- coding: utf8 -*-
'''
Flappy Triangle on 18/11/2014
'''

from FGAme import *
from random import uniform


class Flappy(Poly):

    def __init__(self, **kwds):
        super(Flappy, self).__init__(
            [(0, 0), (40, 0), (20, 80)], color='red', **kwds)
        self.pos = (200, 300)
        self.rotate(uniform(0, 2 * math.pi))
        self.inertia *= 10
        self.omega = uniform(-2, 2)
        self.receiving_input = True

    @listen('key-down', 'space')
    def flappy_up(self):
        '''Aumenta a velocidade vertical do flappy'''

        if self.receiving_input:
            self.boost((0, 150))

    @listen('key-down', 'left', delta=0.2)
    @listen('key-down', 'right', delta=-0.2)
    def change_omega(self, delta):
        '''Modifica a velocidade angular do flappy por um fator delta'''

        if self.receiving_input:
            self.omega += delta

    @listen('collision')
    def block_input(self, col=None):
        '''Bloqueia a entrada do usuário'''

        self.receiving_input = False


class Game(World):

    def __init__(self):
        super(Game, self).__init__(gravity=200)

        # Adiciona obstáculos
        self.N = N = 4
        self.obstacles = []
        for i in range(N):
            self.new_obstacle((850 / N) * (i + 1) + 400)

        # Adiciona o chão e teto
        self.floor = AABB(bbox=(0, 800, -300, 10), mass='inf', world=self)
        self.ceiling = AABB(bbox=(0, 800, 590, 800), mass='inf', world=self)

        # Adiciona o Flappy
        self.flappy = Flappy(world=self)

    def new_obstacle(self, pos_x):
        '''Cria um novo obstáculo na posição pos_x'''

        size = 50
        speed = 50
        middle = uniform(50 + size, 550 - size)
        lower = AABB(bbox=(pos_x, pos_x + 30, 0, middle - size),
                     mass='inf', vel=(-speed, 0), world=self)
        upper = AABB(bbox=(pos_x, pos_x + 30, middle + size, 600),
                     mass='inf', vel=(-speed, 0), world=self)
        self.obstacles.append([lower, upper])

    @listen('frame-enter')
    def detect_exit(self):
        '''Detecta se um obstáculo saiu da tela e insere um novo em caso
        afirmativo'''

        L = self.obstacles
        if L[0][0].xmax < 0:
            self.remove(L[0][0])
            self.remove(L[0][1])
            del L[0]
            new_x = L[-1][0].xmin + 850 / self.N
            self.new_obstacle(new_x)

        if self.flappy.xmax < -400:
            self.game_over()

    def game_over(self):
        '''Game over'''

        self.stop()
        GameOver().run()


class GameOver(World):

    def __init__(self):
        from FGAme.extra.letters import add_word

        super(GameOver, self).__init__(background=(255, 0, 0), gravity=10)
        Rectangle(shape=(600, 100), pos=(0, -200),
                  theta=math.pi / 9, world=self, mass='inf', color=(255, 0, 0))
        letters = add_word('game over', self, scale=5, pos=(-220, 50))
        for l in letters:
            l.inertia *= 20

        self.listen('key-down', 'enter', self.reinit)

    def reinit(self):
        self.stop()
        Game().run()

if __name__ == '__main__':
    game = Game()
    print(game._simulation._objects)
    game.run()



# -*- coding: utf8 -*-_

'''
Este exemplo demonstra o empilhamento de círculos.
'''

from FGAme import World, Circle

# Cria mundo com coeficiente de atrito global não-n'ulo
world = World(friction=0.5, gravity=300, restitution=0.0)
world.add_bounds(width=10)

N = 5
aabbs = []
for i in range(N):
    aabbs.append(Circle(25 + 5 * i,
                        pos=(400, 50 + i * 100), world=world, color='red'))
aabbs[0].make_static()
world.register_energy_tracker()
world.run()



# -*- coding: utf8 -*-

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO
from collections import MutableSequence
import string


class Parser(object):

    '''Processa um arquivo de declaração de objetos. Retorna uma árvore com a
    hierarquia implícita na declaração. A estrutura resultante é mais
    conveniente para processamento futuro.

    >>> example = """
    ... Rect:
    ...     shape: (100, 200)
    ...     vel: (100, 100)
    ...     color: red
    ...
    ... # Comentários
    ... draw.Triangle:
    ...     length: 100
    ...     pos: (0, 1)
    ...     color: black
    ...
    ... Rect:
    ... Triangle:"""
    >>> tree = parse_data(example)
    >>> print(tree[1])
    draw.Triangle:
        length: 100
        pos: (0, 1)
        color: 'black'
    '''

    def __init__(self, buffer):
        if isinstance(buffer, str):
            buffer = StringIO(buffer)
        self.buffer = buffer

    def parse(self):
        '''Retorna uma lista de listas onde cada nó representa um ponto na
        hierarquia definida pela declaração fornecida junto com seu número de
        linha.'''

        lines = list(enumerate(L.rstrip() for L in self.buffer.readlines()))
        root = Root()
        currpath = [root]
        identation = -1

        for ln, line in get_logical_lines(lines):
            idx = get_indentation(line)

            # Linhas que não terminam com ":" são atribuições de valor.
            # Modifica o nó para que estas linhas já carreguem o valor
            # processado
            data = line.strip()
            node = self.process_line(ln, data)

            # Mesmo nível hierárquico
            if idx == identation:
                currpath[-2].children.append(node)
                currpath[-1] = node

            # Sobe um nível hierárquico
            elif idx <= identation:
                currpath.pop()
                currpath[-2].children.append(node)
                currpath[-1] = node

            # Desce um nível hierárquico
            else:
                currpath[-1].children.append(node)
                currpath.append(node)

            identation = idx

        return root

    def process_line(self, ln, data):
        '''Retorna um objeto que representa a instrução desenvolvida na linha
        fornecida'''

        if data.endswith(':'):
            return Header.from_string(data, ln)
        else:
            return Assignment.from_string(data, ln)

#=========================================================================
# Classes auxiliares e nós na árvore de Parse
#=========================================================================
#
# Classe base para criação de nós
#


class Root(MutableSequence):

    def __init__(self):
        self.children = []

    def __str__(self):
        return '\n'.join(str(x) for x in self.children)

    def __repr__(self):
        return 'Root(%r)' % self.children

    # Métodos abstratos de MutableSequence
    def __delitem__(self, idx):
        del self.children[idx]

    def __getitem__(self, idx):
        return self.children[idx]

    def __len__(self):
        return len(self.children)

    def __setitem__(self, idx, value):
        self.children[idx] = value

    def insert(self, idx, value):
        self.children.insert(idx, value)


class Header(object):

    def __init__(self, name, ln=0):
        self.name = name
        self.ln = ln
        self.children = []

    @classmethod
    def from_string(cls, data, ln):
        data = data.strip()
        data = data[:-1]
        if data[0] == '<' and data[-1] == '>':
            return Group(data[1:-1], ln)
        else:
            return Declaration(data, ln)

    def __str__(self):
        out = [self.strheader()]
        for child in self.children:
            for line in str(child).splitlines():
                out.append('    ' + line)
        return '\n'.join(out)


class Group(Header):

    def strheader(self):
        return '<%s>:' % self.name


class Declaration(Header):

    def strheader(self):
        return '%s:' % self.name


class Assignment(object):
    __slots__ = ['name', 'value', 'ln']

    def __init__(self, name, value, ln=0):
        self.name = name
        self.value = value
        self.ln = ln

    def __repr__(self):
        return '%s: %r' % (self.name, self.value)

    @classmethod
    def from_string(cls, data, ln=0):
        '''Inicializa a partir de uma string do tipo "foo: bar"'''

        data = data.strip()
        name, _, data = data.partition(':')
        data = cls._process_data(data)
        return cls(name, data, ln=ln)

    @classmethod
    def _process_data(cls, data):
        # HACK!! Implemente isso direito!
        data = data.strip()

        # Números
        if data[0].isdigit():
            return number(data)

        # Strings
        elif data.isalnum() and data[0] in string.ascii_letters:
            return data
        elif (data[0] == data[-1]) and data[0] in ['"', "'"]:
            return data

        # Tuplas
        elif data[0] == '(' and data[-1] == ')':
            data = data[1:-1].split(',')
            return tuple(cls._process_data(x) for x in data)

        # Listas
        elif data[0] == '[' and data[-1] == ']':
            data = data[1:-1].split(',')
            return [cls._process_data(x) for x in data]

        # Erro
        else:
            raise ValueError('could not process data: %r' % data)

#=========================================================================
# Funções auxiliares
#=========================================================================


def number(data):
    '''Converte uma string em um número'''
    try:
        return int(data)
    except:
        return float(data)


def get_logical_lines(lines):
    '''Processa as linhas eliminando linhas vazias e comentários'''

    out = []
    for i, line in lines:
        # Discard empty lines
        if not line.strip():
            continue

        # Strip comment lines
        if line.lstrip().startswith('#'):
            continue

        out.append((i, line))
    return out


def get_indentation(st):
    '''Retorna um inteiro com o nível de identação da string `st`'''

    idx = 0
    for c in st:
        if c == ' ':
            idx += 1
        else:
            break
    return idx


def parse_data(data):
    '''Processa o conteúdo de declarações de objeto fornecido e retorna uma
    árvore com a hierarquia implícita no arquivo.'''

    return Parser(data).parse()

if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf8 -*-

from . import LinearRigidBody
from FGAme.mathtools import sqrt, shapes

__all__ = ['AABB']


class AABB(LinearRigidBody):

    '''Define um objeto físico que responde a colisões como uma caixa de
    contorno alinhada aos eixos.

    Deve ser inicializada ou por uma tupla com os valores (xmin, xmax, ymin,
    ymax) ou por definido shape=(lado x, lado y).


    Example
    -------

    Os objetos tipo AABB podem ser iniciados de 4 maneiras diferentes.
    A primeira é fornecendo um formato e opcionalmente a posição do centro::

        >>> A = AABB(shape=(100, 200))
        >>> A.bbox  # (xmin, xmax, ymin, ymax)
        (-50.0, 50.0, -100.0, 100.0)

    A segunda opção é indicar o uma tupla com a posição do vértice inferior
    esquerdo e a forma::

        >>> B = AABB(rect=(0, 0, 100, 200))
        >>> B.bbox
        (0.0, 100.0, 0.0, 200.0)

    A outra maneira é fornecer a caixa de contorno como uma lista::

        >>> C = AABB(bbox=(0, 100, 0, 200))

    E, finalmente, podemos fornecer os valores de xmin, xmax, ymin, ymax
    individualmente::

        >>> D = AABB(0, 100, 0, 200)
        >>> E = AABB(xmin=0, xmax=100, ymin=0, ymax=200)

    Em todos os casos, a massa já é calculada automaticamente a partir da área
    assumindo-se uma densidade de 1.0. O momento de inércia é infinito pois
    trata-se de um objeto sem dinâmica angular::

        >>> A.mass, A.inertia  # 20.000 == 200 x 100
        (20000.0, inf)

    '''

    __slots__ = []

    def __init__(self, xmin=None, xmax=None, ymin=None, ymax=None,
                 pos=None, vel=(0, 0), mass=None, density=None,
                 bbox=None, shape=None, rect=None, **kwds):

        # Define as propriedades das caixas de contorno
        xmin, xmax, ymin, ymax = shapes.aabb_bbox(bbox=bbox, rect=rect,
                                                  shape=shape, pos=pos,
                                                  xmin=xmin, xmax=xmax,
                                                  ymin=ymin, ymax=ymax)

        pos = ((xmin + xmax) / 2., (ymin + ymax) / 2.)
        self._delta_x = dx = (xmax - xmin) / 2
        self._delta_y = dy = (ymax - ymin) / 2
        self.cbb_radius = sqrt(dx ** 2 + dy ** 2)
        aabb = shapes.AABB(-dx, dx, -dy, dy)
        super(AABB, self).__init__(pos, vel, mass=mass, density=density,
                                   baseshape=aabb, **kwds)

    def __repr__(self):
        tname = type(self).__name__
        if not self._invmass:
            tname += '*'
        vel = ', '.join('%.1f' % x for x in self._vel)
        data = ', '.join('%.1f' % x for x in self.bbox)
        return '%s(bbox=[%s], vel=(%s))' % (tname, data, vel)

    # Torna as os limites da AABB mutáveis ####################################
    @property
    def aabb(self):
        return self.bounding_box

    @property
    def xmin(self):
        return self._pos.x - self._delta_x

    @xmin.setter
    def xmin(self, value):
        xmin = float(value)
        xmax = self.xmax
        self._pos = self._pos.setx((xmax + xmin) / 2)
        self._delta_x = (xmax - xmin) / 2

    @property
    def xmax(self):
        return self._pos.x + self._delta_x

    @xmax.setter
    def xmax(self, value):
        xmin = self.xmin
        xmax = float(value)
        self._pos = self._pos.setx((xmax + xmin) / 2)
        self._delta_x = (xmax - xmin) / 2

    @property
    def ymin(self):
        return self._pos.y - self._delta_y

    @ymin.setter
    def ymin(self, value):
        ymin = float(value)
        ymax = self.ymax
        self._pos = self._pos.sety((ymax + ymin) / 2)
        self._delta_y = (ymax - ymin) / 2

    @property
    def ymax(self):
        return self._pos.y + self._delta_y

    @ymax.setter
    def ymax(self, value):
        ymin = self.ymin
        ymax = float(value)
        self.pos = self.pos.copy(y=(ymax + ymin) / 2)
        self._delta_y = (ymax - ymin) / 2

    # Propriedades geométricas ################################################
    def area(self):
        return 4 * self._delta_x * self._delta_y

    def ROG_sqr(self):
        a = self._delta_x
        b = self._delta_y
        return (a ** 2 + b ** 2) / 3


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*-coding: utf8 -*-
import random
from math import sqrt, acos


class Color(object):

    '''
    Objeto básico que representa uma cor.

    O construtor de Color() aceita várias assinaturas de entrada, que
    representam maneiras diferentes de construir uma cor. A maneira mais
    simples e direta consiste em especificar as componentes RGB ou RGBA.

    >>> w1 = Color(255, 255, 255)
    >>> w2 = Color(255, 255, 255, 255)
    >>> w1 == w2
    True

    Além disto, podemos especificar a cor pelo nome.

    >>> Color('white')
    Color(255, 255, 255, 255)

    Todos os nomes do padrão CSS3.0 são aceitos. A lista completa de cores
    pode ser acessada em http://www.w3.org/TR/css3-color/#svg-color.

    Também é possível especificar a cor pelo código hexadecimal. Isto pode ser
    feito através de uma string ou utilizando a sintaxe do Python para gerar
    inteiros a partir da representação hexadecimal.

    >>> red = Color('#F00')
    >>> lime = Color('#00FF00')
    >>> blue = Color(0x0000ff, alpha=128)

    Nos dois primeiros casos o canal alpha é opcional e pode ser embutido na
    string com o código hex. No segundo caso, é necessário passar o canal alpha
    como argumento opcional. O valor padrão é 255, que corresponde a uma cor
    sólida.

    Note que apesar de RGB significar *Red* (vermelho), *Green* (verde), *Blue*
    (azul), o nome web para a cor com o canal verde puro é *Lime*. A cor
    "green" corresponde ao código hex #008000. A FGAme utiliza a mesma
    convenção.


    Examples
    --------

    Os objetos do tipo Color são imutáveis e se comportam como uma tupla.

    >>> list(w1)
    [255, 255, 255, 255]
    >>> w1
    Color(255, 255, 255, 255)

    Podemos acessar a cor em várias representações diferentes utilizando os
    atributos adequados.

    >>> w1.rgb
    (255, 255, 255)

    >>> w1.f_rgb
    (1.0, 1.0, 1.0)

    >>> w1.u_rgb
    16777215

    '''

    __slots__ = ['_red', '_green', '_blue', '_alpha']
    _CACHE = {}
    _RANGE = (0, 1, 2, 3)

    def __new__(cls, value, *args, alpha=255):
        if not args:
            if isinstance(value, Color):
                return value
            elif isinstance(value, tuple):
                A = alpha
                if len(value) == 3:
                    R, G, B = value
                else:
                    R, G, B, A = value
            elif isinstance(value, str):
                value = value.lower()
                if value.startswith('#'):
                    value = value[1:]
                    A = alpha
                    if len(value) == 6:
                        R, G, B = (
                            int(value[2 * i:2 * i + 2], 16) for i in range(3))
                    elif len(value) == 8:
                        R, G, B, A = (
                            int(value[2 * i:2 * i + 2], 16) for i in range(4))
                    elif 3 <= len(value) <= 4:
                        data = (255 * int(x, 16) / 15 for x in value)
                        if len(value) == 3:
                            R, G, B = data
                        else:
                            R, G, B, A = data
                    else:
                        raise ValueError('invalid hex code: #%s' % value)
                elif value == 'random':
                    A = 255
                    R, G, B = [random.randint(0, 255) for _ in range(3)]
                else:
                    return cls._CACHE[value]
            elif isinstance(value, int):
                R = (value & 0xff0000) >> 16
                G = (value & 0xff00) >> 8
                B = (value & 0xff) >> 0
                A = alpha
            else:
                raise ValueError
        else:
            R = value
            A = alpha
            if len(args) == 2:
                G, B = args
            else:
                G, B, A = args

        # Create new object
        new = object.__new__(cls)
        new._red = int(R)
        new._green = int(G)
        new._blue = int(B)
        new._alpha = int(A)
        return new

    # Componentes RGBa
    @property
    def red(self):
        return self._red

    @property
    def green(self):
        return self._green

    @property
    def blue(self):
        return self._blue

    @property
    def alpha(self):
        return self._alpha

    # Componentes HSI
    @property
    def hue(self):
        return NotImplemented

    @property
    def saturation(self):
        return NotImplemented

    @property
    def intensity(self):
        return NotImplemented

    # Representações em diferentes espaço de cores
    @property
    def rgba(self):
        return tuple(self)

    @property
    def rgb(self):
        return self[:3]

    @property
    def f_rgb(self):
        return tuple(x / 255. for x in self[:3])

    @property
    def f_rgba(self):
        return tuple(x / 255. for x in self)

    @property
    def u_rgba(self):
        c = self
        return (c[0] << 24) + (c[1] << 16) + (c[2] << 8) + c[3]

    @property
    def u_rgb(self):
        c = self
        return (c[0] << 16) + (c[1] << 8) + c[2]

    @property
    def hsi(self):
        return self.hsia[:-1]

    @property
    def hsia(self):
        # Conversão para (H)ue, (S)aturation, (I)ntensity
        # ref: https://en.wikipedia.org/wiki/RGB_color_model#Nonlinearity
        R, G, B, a = self

        I = (R + G + B) / 3
        S = 1 - min(R, G, B) / I

        # Valores normalizados
        r, g, b = R / 255, G / 255, B / 255
        h_numer = acos(((r - g) + (r - b)) / 2)
        h_denom = sqrt((r - b) ** 2 + (r - b) * (g - b))
        return h_numer / h_denom, S, I, a

    @property
    def f_hsia(self):
        return tuple(x / 255 for x in self.hsia)

    @property
    def f_hsi(self):
        return tuple(x / 255 for x in self.hsi)

    def copy(self, red=None, green=None, blue=None, alpha=None, **kwds):
        '''Copia a cor possivelmente trocando o valor de alguma das componentes
        RGBA ou componentes HSL ou HSI.

        Examples
        --------

        >>> color = Color('white')
        >>> color.copy(red=80, alpha=128)
        Color(80, 255, 255, 128)
        '''
        R, G, B, A = self

        if red is not None:
            R = red
        if green is not None:
            G = green
        if blue is not None:
            B = blue
        if alpha is not None:
            A = alpha
        if kwds:
            raise NotImplementedError

        return Color(R, G, B, A)

    # Métodos mágicos #########################################################
    def __repr__(self):
        return 'Color%s' % (tuple(self),)

    def __len__(self):
        return 4

    def __eq__(self, other):
        try:
            return (
                (self._red == other._red) and
                (self._green == other._green) and
                (self._blue == other._blue) and
                (self._alpha == other._alpha))
        except (AttributeError, TypeError):
            if len(self) == len(other):
                return all(x == y for (x, y) in zip(self, other))
            return False

    def __iter__(self):
        yield self._red
        yield self._green
        yield self._blue
        yield self._alpha

    def __getitem__(self, key):
        if isinstance(key, int):
            if key == 0:
                return self._red
            elif key == 1:
                return self._green
            elif key == 2:
                return self._blue
            elif key == 3:
                return self._alpha
            elif key < 0:
                key = len(self) - key
                if key < 0:
                    raise IndexError
                return self[key]
        else:
            return tuple(self[i] for i in self._RANGE[key])
        raise IndexError(key)

    def __hash__(self):
        return hash(self.u_rgba)

Color._CACHE.update(
    # HTML 4.01 colors.
    # See https://en.wikipedia.org/wiki/Web_colors
    white=Color('#ffffff'),
    silver=Color('#c0c0c0'),
    gray=Color('#808080'),
    black=Color('#000000'),
    red=Color('#ff0000'),
    maroon=Color('#800000'),
    yellow=Color('#ffff00'),
    olive=Color('#808000'),
    lime=Color('#00ff00'),
    green=Color('#008000'),
    aqua=Color('#00ffff'),
    teal=Color('#008080'),
    blue=Color('#0000ff'),
    navy=Color('#000080'),
    fuschia=Color('#ff00ff'),
    purple=Color('#800080'),
    transparent=Color('#ffffff00'),
    null=Color('#00000000'),
)

# SVG extended Colors
# http://www.w3.org/TR/css3-color/#svg-color
COLOR_NAMES = [
    ['aliceblue', '#F0F8FF'],
    ['antiquewhite', '#FAEBD7'],
    ['aqua', '#00FFFF'],
    ['aquamarine', '#7FFFD4'],
    ['azure', '#F0FFFF'],
    ['beige', '#F5F5DC'],
    ['bisque', '#FFE4C4'],
    ['black', '#000000'],
    ['blanchedalmond', '#FFEBCD'],
    ['blue', '#0000FF'],
    ['blueviolet', '#8A2BE2'],
    ['brown', '#A52A2A'],
    ['burlywood', '#DEB887'],
    ['cadetblue', '#5F9EA0'],
    ['chartreuse', '#7FFF00'],
    ['chocolate', '#D2691E'],
    ['coral', '#FF7F50'],
    ['cornflowerblue', '#6495ED'],
    ['cornsilk', '#FFF8DC'],
    ['crimson', '#DC143C'],
    ['cyan', '#00FFFF'],
    ['darkblue', '#00008B'],
    ['darkcyan', '#008B8B'],
    ['darkgoldenrod', '#B8860B'],
    ['darkgray', '#A9A9A9'],
    ['darkgreen', '#006400'],
    ['darkgrey', '#A9A9A9'],
    ['darkkhaki', '#BDB76B'],
    ['darkmagenta', '#8B008B'],
    ['darkolivegreen', '#556B2F'],
    ['darkorange', '#FF8C00'],
    ['darkorchid', '#9932CC'],
    ['darkred', '#8B0000'],
    ['darksalmon', '#E9967A'],
    ['darkseagreen', '#8FBC8F'],
    ['darkslateblue', '#483D8B'],
    ['darkslategray', '#2F4F4F'],
    ['darkslategrey', '#2F4F4F'],
    ['darkturquoise', '#00CED1'],
    ['darkviolet', '#9400D3'],
    ['deeppink', '#FF1493'],
    ['deepskyblue', '#00BFFF'],
    ['dimgray', '#696969'],
    ['dimgrey', '#696969'],
    ['dodgerblue', '#1E90FF'],
    ['firebrick', '#B22222'],
    ['floralwhite', '#FFFAF0'],
    ['forestgreen', '#228B22'],
    ['fuchsia', '#FF00FF'],
    ['gainsboro', '#DCDCDC'],
    ['ghostwhite', '#F8F8FF'],
    ['gold', '#FFD700'],
    ['goldenrod', '#DAA520'],
    ['gray', '#808080'],
    ['green', '#008000'],
    ['greenyellow', '#ADFF2F'],
    ['grey', '#808080'],
    ['honeydew', '#F0FFF0'],
    ['hotpink', '#FF69B4'],
    ['indianred', '#CD5C5C'],
    ['indigo', '#4B0082'],
    ['ivory', '#FFFFF0'],
    ['khaki', '#F0E68C'],
    ['lavender', '#E6E6FA'],
    ['lavenderblush', '#FFF0F5'],
    ['lawngreen', '#7CFC00'],
    ['lemonchiffon', '#FFFACD'],
    ['lightblue', '#ADD8E6'],
    ['lightcoral', '#F08080'],
    ['lightcyan', '#E0FFFF'],
    ['lightgoldenrodyellow', '#FAFAD2'],
    ['lightgray', '#D3D3D3'],
    ['lightgreen', '#90EE90'],
    ['lightgrey', '#D3D3D3'],
    ['lightpink', '#FFB6C1'],
    ['lightsalmon', '#FFA07A'],
    ['lightseagreen', '#20B2AA'],
    ['lightskyblue', '#87CEFA'],
    ['lightslategray', '#778899'],
    ['lightslategrey', '#778899'],
    ['lightsteelblue', '#B0C4DE'],
    ['lightyellow', '#FFFFE0'],
    ['lime', '#00FF00'],
    ['limegreen', '#32CD32'],
    ['linen', '#FAF0E6'],
    ['magenta', '#FF00FF'],
    ['maroon', '#800000'],
    ['mediumaquamarine', '#66CDAA'],
    ['mediumblue', '#0000CD'],
    ['mediumorchid', '#BA55D3'],
    ['mediumpurple', '#9370DB'],
    ['mediumseagreen', '#3CB371'],
    ['mediumslateblue', '#7B68EE'],
    ['mediumspringgreen', '#00FA9A'],
    ['mediumturquoise', '#48D1CC'],
    ['mediumvioletred', '#C71585'],
    ['midnightblue', '#191970'],
    ['mintcream', '#F5FFFA'],
    ['mistyrose', '#FFE4E1'],
    ['moccasin', '#FFE4B5'],
    ['navajowhite', '#FFDEAD'],
    ['navy', '#000080'],
    ['oldlace', '#FDF5E6'],
    ['olive', '#808000'],
    ['olivedrab', '#6B8E23'],
    ['orange', '#FFA500'],
    ['orangered', '#FF4500'],
    ['orchid', '#DA70D6'],
    ['palegoldenrod', '#EEE8AA'],
    ['palegreen', '#98FB98'],
    ['paleturquoise', '#AFEEEE'],
    ['palevioletred', '#DB7093'],
    ['papayawhip', '#FFEFD5'],
    ['peachpuff', '#FFDAB9'],
    ['peru', '#CD853F'],
    ['pink', '#FFC0CB'],
    ['plum', '#DDA0DD'],
    ['powderblue', '#B0E0E6'],
    ['purple', '#800080'],
    ['red', '#FF0000'],
    ['rosybrown', '#BC8F8F'],
    ['royalblue', '#4169E1'],
    ['saddlebrown', '#8B4513'],
    ['salmon', '#FA8072'],
    ['sandybrown', '#F4A460'],
    ['seagreen', '#2E8B57'],
    ['seashell', '#FFF5EE'],
    ['sienna', '#A0522D'],
    ['silver', '#C0C0C0'],
    ['skyblue', '#87CEEB'],
    ['slateblue', '#6A5ACD'],
    ['slategray', '#708090'],
    ['slategrey', '#708090'],
    ['snow', '#FFFAFA'],
    ['springgreen', '#00FF7F'],
    ['steelblue', '#4682B4'],
    ['tan', '#D2B48C'],
    ['teal', '#008080'],
    ['thistle', '#D8BFD8'],
    ['tomato', '#FF6347'],
    ['turquoise', '#40E0D0'],
    ['violet', '#EE82EE'],
    ['wheat', '#F5DEB3'],
    ['white', '#FFFFFF'],
    ['whitesmoke', '#F5F5F5'],
    ['yellow', '#FFFF00'],
    ['yellowgreen', '#9ACD32']
]
for _name, _code in COLOR_NAMES:
    Color._CACHE.setdefault(_name, Color(_code))

#
# Funções e objetos úteis
#


class color_property(property):

    '''
    Propriedade que converte automaticamente os valores fornecidos como
    strings, tuplas ou inteiros em objetos do tipo Color.

    Mantêm o valor None, caso fornecido.
    '''

    def __init__(self, name, default=None):
        self.name = name
        self.default = (None if default is None else Color(default))
        attr = '_' + name

        def fget(self):
            return getattr(self, attr, default)

        def fset(self, value):
            if value is None:
                fdel(self)
            else:
                setattr(self, attr, Color(value))

        def fdel(self):
            if hasattr(self, attr):
                delattr(self, attr)

        super(color_property, self).__init__(fget, fset, fdel)


def rgb(color):
    '''Convert input in a tuple of (red, green, blue) colors.

    Null values are converted to solid black.'''

    try:
        return color.rgb
    except AttributeError:
        return Color(color or 'black').rgb


def rgba(color):
    '''Convert input in a tuple of (red, green, blue, alpha) components

    Null values are converted to solid black.'''

    try:
        return color.rgba
    except AttributeError:
        return Color(color or 'black').rgba

if __name__ == '__main__':
    import doctest
    doctest.testmod()

    print(Color('white'), Color(255, 255, 255), Color(0xffffffff))
    assert Color('white') == Color(255, 255, 255)
    assert Color('white') == Color(255, 255, 255, 255)
    for hexcode in ['#FFF', '#FFFF', '#FFFFFF', '#FFFFFFFF']:
        assert Color('white') == Color(hexcode)
    assert Color('white') == Color(0xffffff)



'''
Loads all tests in module and run
'''

from smallvectors.tests import *
from FGAme.tests import *
import FGAme as mod_current

from unittest2 import main
import doctest
import sys
from FGAme import conf
conf.set_backend()
print('Starting tests using backend: %s' % conf.get_backend())


def load_tests(loader, tests, ignore):
    prefix = mod_current.__name__

    # Find doctests
    for modname, mod in sys.modules.items():
        if modname.startswith(prefix + '.') or modname == prefix:
            try:
                tests.addTests(doctest.DocTestSuite(mod))
            except ValueError:  # no docstring
                pass

    return tests

main()



# -*- coding: utf8 -*-
'''
Classes abstratas
-----------------

Todos os objetos físicos na FGAme herdam da classe Dynamic ou de uma das
subclasses derivadas aqui. Não é possível instanciar estas classes diretamente,
mas elas são a base para todos os tipos de objetos utilizados na FGAme.

.. automodule:: FGAme.objects2.base

Classes derivadas
-----------------

#.. autoclass:: FGAme.AABB
#.. autoclass:: FGAme.Circle
#.. autoclass:: FGAme.Poly
#.. autoclass:: FGAme.Letter

Funções especiais para a criação de objetos
--------------------------------------------

#.. automodule :: FGAme.letters

'''

from .util import *
from .flags import *
from .bodies import *
from . import flags
from .collision import Collision, CBBContact, AABBContact
from .collision_pairs import get_collision
from . import collision_pairs as collision
from .forces import *
from .simulation import *


__all__ = ['Collision', 'get_collision', 'Simulation']



# -*- coding: utf8 -*-

import os
import importlib

__all__ = ['get_screen_class', 'get_input_class', 'get_mainloop_class',
           'get_backend_classes', 'supports_backend']
GLOBAL_INFO = None


def get_info():
    '''Retorna um dicionário mapeando cada backend com um dicionário que
    mapeia 'input', 'screen' e 'mainloop' nos nomes das respectivas classes'''

    global GLOBAL_INFO

    if GLOBAL_INFO is not None:
        return GLOBAL_INFO

    else:
        GLOBAL_INFO = out = {}
        path, _ = os.path.split(__file__)
        for path in os.listdir(path):
            if path.endswith('_conf.py'):
                name, _, _ = path.rpartition('_')
                conf = importlib.import_module('FGAme.backends.%s_conf' % name)
                out[name] = dict(
                    input=conf.input,
                    screen=conf.screen,
                    mainloop=conf.mainloop,
                    imports=conf.imports,
                )
        return out


def get_backend_classes(backend):
    '''Retorna um dicionário mapeando 'input', 'screen' e 'mainloop' nas
    respectivas classes para o backend selecionado.

    Exemplos
    --------

    >>> from FGAme.backends import pygame_be
    >>> D = get_backend_classes('pygame')
    >>> D == {'input':    pygame_be.PyGameInput,
    ...       'mainloop': pygame_be.PyGameMainLoop,
    ...       'screen':   pygame_be.PyGameCanvas}
    True
    '''

    D = get_info()[backend]
    classes = ('input', 'mainloop', 'screen')
    return {k: _get_class_worker(k, backend) for k in classes}


def _get_class_worker(cls, backend):
    '''Implementa as funções get_screen, get_input, get_mainloop, etc'''

    D = get_info()[backend]
    module = importlib.import_module('FGAme.backends.%s_be' % backend)
    return getattr(module, D[cls])


def get_screen_class(backend):
    '''Retorna a classe de screen para o backend selecionado'''

    return _get_class_worker('screen', backend)


def get_input_class(backend):
    '''Retorna a classe de input para o backend selecionado'''

    return _get_class_worker('input', backend)


def get_mainloop_class(backend):
    '''Retorna a classe de mainloop para o backend selecionado'''

    return _get_class_worker('mainloop', backend)


def supports_backend(backend):
    '''Retorna True caso o backend fornecido seja suportado'''

    imports = get_info()[backend]['imports']
    try:
        for module in imports:
            importlib.import_module(module)
        else:
            return True
    except ImportError:
        return False

if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf8 -*-
'''
Este exemplo mostra um gás de esferas rígidas em um campo gravitacional.
'''

from FGAme import World, Circle, AABB, Vec2, listen
from random import uniform


class Gas(World):

    def __init__(self,
                 gravity=100, friction=0.0, restitution=0.95,
                 num_balls=80, speed=200, radius=10,
                 color='random'):
        '''Cria uma simulação de um gás de partículas sujeito à ação da
        gravidade aberto na parte de cima.

        O parâmetros `num_balls` controla o número esferas de raio ``radius``
        com velocidades no intervalo de +/-`speed`.'''

        self.num_balls = num_balls
        self.radius = radius
        self.speed = speed

        super(Gas, self).__init__(gravity=gravity, friction=friction,
                                  restitution=restitution)

    def init(self):
        # Inicia paredes
        AABB(0, 30, 30, 800, world=self, mass='inf')
        AABB(770, 800, 30, 800, world=self, mass='inf')
        AABB(0, 800, 0, 30, world=self, mass='inf')

        # Cria lista de bolas
        speed = self.speed
        for _ in range(self.num_balls):
            Circle(
                radius=self.radius,
                vel=Vec2(uniform(-speed, speed), uniform(-speed, speed)),
                pos=Vec2(uniform(50, 750), uniform(50, 400)),
                color='random',
                world=self,
            )

    @listen('key-down', 'space')
    def toggle_pause(self):
        super(Gas, self).toggle_pause()

# Inicia a simulação
if __name__ == '__main__':
    w = Gas()
    w.run()



'''
Created on 01/07/2015

@author: chips
'''
# TODO: mudar run() para start() e fazer o run() iniciar um shell IPython

import functools as _functools
from collections import deque as _deque
import threading as _threading
import smallshapes as _mathtools
from FGAme.mathtools import Vec2
from FGAme import World as _World
from FGAme.draw import Pen as _Pen, Color
from FGAme import draw
import FGAme

_TURTLE = None
_PEN_LOCK = _threading.RLock()
_FGAME_THREAD = None
_RUNNING = False


class Turtle(_Pen):
    _blacklist = [
        'draw',
        'current', 'tip', 'move_tip',
        'lock_pen_z', 'lock_pen_xy'
    ]

    def __init__(self, pos=(0, 0)):
        super(Turtle, self).__init__(pos)
        self._jobs = _deque()

    def set_pos(self, x_or_pos, y=None):
        '''Move o cursor para a posição fornecida'''

        pos = _mathtools.util.args_to_vec2(x_or_pos, y)
        self.pos = pos

    def set_angle(self, value):
        '''Define a orientação do cursor (em graus)'''

        self.angle = value

    def set_theta(self, value):
        '''Define a orientação do cursor (em radianos)'''

        self.theta = value

    def center(self, penup=True):
        '''Move o cursor para o centro'''

        if penup:
            self.up()
        self.pos = FGAme.pos.middle

    #
    # Suporte à animação
    #
    def _add_job(self, func, *args, **kwds):
        with _PEN_LOCK:
            self._jobs.append([(func, args, kwds)])

    def _add_job_first(self, func, *args, **kwds):
        with _PEN_LOCK:
            self._jobs.insert(0, [(func, args, kwds)])

    def _add_job_first_mapping(self, func, seq):
        with _PEN_LOCK:
            L = [[(func, (x,), {})] for x in seq]
            self._jobs.extendleft(L)

    def _add_task(self, func, *args, **kwds):
        with _PEN_LOCK:
            if not self._jobs:
                self._jobs.append([])
            self._jobs[-1].append((func, args, kwds))

    def job_wrap(self, method):
        @_functools.wraps(method)
        def decorated(*args, **kwds):
            self._add_job(method, *args, **kwds)
        return decorated

    def step(self):
        '''Avança uma tarefa de execução'''

        with _PEN_LOCK:
            if self._jobs:
                print(len(self._jobs))
                job = self._jobs.popleft()
                for task in job:
                    task[0](*task[1], **task[2])

    def move_tip(self, pos):
        '''Move a ponta para a nova posição pos'''

        self._pos = self._current[-1] = Vec2(*pos)

    def make_move_job(self, method):
        @_functools.wraps(method)
        def decorated(*args, **kwds):
            pos = self.pos
            method(*args, **kwds)
            final = self.pos
            self._current[-1] = pos
            interpolated = self.interpolate(pos, final, 1)
            self._add_job_first_mapping(self.move_tip, interpolated)
        return decorated

    def interpolate(self, start, end, steps):
        '''Retorna a lista de interpolação que vai de start até end no número
        dado de passos (steps). O ponto inicial não é incluido'''

        delta = (end - start) / float(steps)
        out = [start + (n + 1) * delta for n in range(steps)]
        out[-1] = end  # diminui acúmulo de erros
        return out

    #
    # Renderização
    #
    def current(self):
        '''Retorna um objeto Path com o caminho atual'''

        if self._pen_down and len(self._current) >= 2:
            return draw.Path(self._current, **self._curve_kwds())

    def tip(self):
        '''Retorna um triângulo apontando para a direção da tartaruga'''

        L = 12
        vertices = [(L, 0), (0, L / 3), (0, -L / 3)]
        tip = draw.Poly(vertices, color=Color('red'))
        tip.rotate(self.theta)
        tip.move(self.pos)
        return tip

    def draw(self, canvas):
        super(Turtle, self).draw(canvas)
        path = self.current()
        if path is not None:
            path.draw(canvas)
        self.tip().draw(canvas)

    #
    # Inicialização do módulo
    #
    @classmethod
    def _populate_ns(cls, ns):
        '''Populate the given namespace with turtle functions and a Turtle()
        instance named _TURTLE.'''

        turtle = ns['_TURTLE'] = cls(FGAme.pos.middle)
        for name in dir(turtle):
            method = getattr(turtle, name)

            # Não adiciona implementações manuais de add_job
            if name.endswith('_job'):
                continue
            if name.startswith('_') or not callable(method):
                continue

            # Embrulha de forma padronizada ou adiciona implementação manual
            if name not in cls._blacklist:
                method = getattr(turtle, name + '_job', method)
                ns.setdefault(name, turtle.job_wrap(method))


def start():
    '''Inicia a janela do FGAme turtle ao fundo e espera por eventos'''

    global _RUNNING, _FGAME_THREAD

    if not _RUNNING:
        world = _World()
        world.add(_TURTLE)

        @world.listen('frame-enter')
        def step():
            _TURTLE.step()

        # Roda FGAme num thread secundário
        _FGAME_THREAD = _threading.Thread(target=world.run)
        _FGAME_THREAD.start()
        _RUNNING = True


def run(shell='ipython'):
    '''Inicia o turtle e executa um terminal interativo para interagir com a
    aplicação'''

    start()

    if shell == 'ipython':
        import IPython
        IPython.embed()
    elif shell == 'bpython':
        import bpython
        print('embeding bpython shell')
        bpython.embed()
    else:
        raise ValueError('invalid shell: %s' % shell)


#
# Adiciona métodos do objeto Turtle ao namespace do módulo
#
# TODO: fazer isto manualmente para cada classe de métodos?
Turtle._populate_ns(globals())


#
# Anima funções de movimento linear do cursor
#
def _animate_move_factory(method):
    '''Embrulha funções que implicam em movimento da tartaruga para que o mesmo
    seja animado ao longo de vários frames'''

    turtle = _TURTLE
    move_method = getattr(turtle, method)

    @_functools.wraps(move_method)
    def animated(*args, **kwds):
        start()

        def job():
            pos = turtle.pos
            move_method(*args, **kwds)
            final = turtle.pos
            turtle._current[-1] = pos
            turtle._pos = pos
            steps = turtle.interpolate(pos, final, 10)
            jobs = [[(turtle.move_tip, (x,), {})] for x in steps]
            with _PEN_LOCK:
                turtle._jobs.extendleft(reversed(jobs))

        turtle._add_job(job)

    return animated

G = globals()
for name in ['fwd', 'forward', 'back', 'backwards', 'move', 'goto']:
    G[name] = _animate_move_factory(name)
del G, name


#
# Anima funções de movimento rotacional do cursor
#
def _animate_rotation_factory(method):
    '''Embrulha funções que implicam em movimento rotacional do cursor
    para que o mesmo seja animado ao longo de vários frames'''

    turtle = _TURTLE
    rotate_method = getattr(turtle, method)

    @_functools.wraps(rotate_method)
    def animated(*args, **kwds):
        start()

        def job():
            angle = turtle.angle
            rotate_method(*args, **kwds)
            final = turtle.angle
            steps = turtle.interpolate(angle, final, 10)
            jobs = [[(turtle.set_angle, (x,), {})] for x in steps]
            with _PEN_LOCK:
                turtle._jobs.extendleft(reversed(jobs))

        turtle._add_job(job)

    return animated

G = globals()
for name in ['left', 'right', 'set_angle', 'set_theta']:
    G[name] = _animate_rotation_factory(name)
del G, name


#
# Figuras geométricas adicionais
#
def star(num_sides=7, size=100):
    '''Desenha uma estrela de num_sides lados com o tamanho de cada lado igual
    à size'''

    for _ in range(12):
        fwd(200)  # @UndefinedVariable
        left(7 * 360 / 12)  # @UndefinedVariable

if __name__ == '__main__':
    import IPython
    start()
    star()
    # run()
    print('starting debugging')
    #IPython.embed(header='FGAme Turtle')



# -*- coding: utf8 -*-

from FGAme.mathtools import null2D, shapes
from . import Body

__all__ = ['Circle']


class Circle(Body):

    '''Define um círculo e implementa a detecção de colisão comparando a
    distância entre os centros com a soma dos raios.


    Examples
    --------

    Os círculos devem ser inicializados fornecendo o raio e opcionalmente a
    posição, velocidade e massa ou densidade.

    >>> c1 = Circle(10, (10, 0))     # raio 10 e centro em (10, 10)
    >>> c2 = Circle(10, density=2)   # raio 10 e densidade de 2
    >>> c2.area()
    314.1592653589793
    '''

    __slots__ = []

    def __init__(self, radius, pos=(0, 0), vel=(0, 0),
                 mass=None, density=None, **kwds):

        self.cbb_radius = float(radius)
        super(Circle, self).__init__(
            pos, vel, mass=mass, density=density,
            baseshape=shapes.Circle(self.cbb_radius, null2D), ** kwds)

    @property
    def radius(self):
        return self.cbb_radius

    @radius.setter
    def radius(self, value):
        self.cbb_radius = value

    def rescale(self, scale, update_physics=False):
        self.cbb_radius *= scale
        super(Circle, self).rescale(scale, update_physics)

    # Caixa de contorno #######################################################
    @property
    def bounding_box(self):
        return self.cbb

    @property
    def xmin(self):
        return self._pos.x - self.cbb_radius

    @property
    def xmax(self):
        return self._pos.x + self.cbb_radius

    @property
    def ymin(self):
        return self._pos.y - self.cbb_radius

    @property
    def ymax(self):
        return self._pos.y + self.cbb_radius

    def __repr__(self):
        tname = type(self).__name__
        vel = ', '.join('%.1f' % x for x in self._vel)
        pos = ', '.join('%.1f' % x for x in self._pos)
        return '%s(pos=(%s), vel=(%s), radius=%.1f)' % (
            tname, pos, vel, self.radius)


if __name__ == '__main__':
    from doctest import testmod
    testmod()



# -*- coding: utf-8 -*-
'''
O controle de despacho de mensagens e eventos é centralizado na classe
EventDispatcher deste módulo. Ela controla a transmissão de "mensagens" ou
eventos entre objetos implementado o padrão "observador", onde um objeto pode
anunciar eventos e outros objetos podem se registrar a estes eventos e
responder apropriadamente quando estes eventos forem disparados.

Este mecanismo é utilizado para desacoplar o código de diferentes partes do
motor de jogos e é essencial para manter o código organizado e compreensivel.
Um exemplo típico da utilidade do despachante de eventos se dá na implementação
da resposta às entradas do usuário. Uma parte do motor de jogos deve ser
dedicada a ler os comandos de teclado, mouse, etc. Ao invés de responder
explicitamente à cada interação, o sistema de entrada simplesmente anuncia algo
do tipo: *a tecla "p" foi pressionada*. Todos os objetos que estiverem
registrados para ouvir este evento serão notificados (e assim chamando a função
de callback() apropriada).

O sinais associados a cada objeto são definidos durante a criação da classe.
Vamos começar criando uma classe que define alguns sinais simples.

>>> class Foobar(EventDispatcher):
...     foo = signal('foo')
...     bar = signal('bar')

Um objeto da classe Foobar() pode disparar os sinais apropriados utilizando
os métodos `.trigger_foo()` ou `.trigger_bar()` ou simplesmente o método
genérico `.trigger('nome')`

>>> x = Foobar()
>>> x.trigger_foo()     # aciona o sinal foo
>>> x.trigger('bar')    # aciona o sinal bar

Os métodos `trigger_<nome do sinal>()` são ligeiramente mais eficientes e
são criados automaticamente pelo construtor das classes do tipo
EventDispatcher. Sub-classes podem personalizar esses métodos, mas é
potencialmente complicado difícil e raramente é necessário.

`x.trigger(...)` por enquanto não produz nenhum efeito pois não há objetos
ouvindo algum sinal de x. Podemos registrar o interesse em um determinado sinal
utilizando o método `x.listen()`, passando uma função que toma uma ação de
resposta a esses sinais.

>>> def handler(*args, **kwds):
...     print('called with args=%s, kwargs=%s' % (args, kwds))

Novamente, aqui existem duas interfaces análogas ao método trigger*()

>>> id1 = x.listen_foo(handler)
>>> id2 = x.listen('bar', handler)
...

Assim como antes, `.listen(nome, ...)` simplesmente chama a função
apropriada `.listen_nome(...)`. Isto pode ser utilizado em sub-classes para
personalizar o comportamento de como cada sinal é registrado ou executado.

O sinais disparados por `x` agora invocam a função `handler` definida
anteriormente.

>>> x.trigger_bar()
called with args=(), kwargs={}

Também podemos registrar as funções com argumentos adicionais que serão
repassados quando elas forem invocadas pelo disparo de um sinal.

>>> id3 = x.listen('foo', handler, 'pos_arg', kw_arg=None)

Primeiramente x.trigger_foo() invocará a função "handler" sem argumentos (como
acontecia anteriormente) e em seguida executará esta mesma função com os dois
argumentos adicionais registrados no bloco acima. Deste modo,

>>> x.trigger_foo()
called with args=(), kwargs={}
called with args=('pos_arg',), kwargs={'kw_arg': None}

Finalmente, o objeto x.foo (ou, analogamente, x.bar) permite controlar
explicitamente as funções registradas sob cada sinal e algumas
características de seu comportamento.

>>> x.foo.handlers                                         # doctest: +ELLIPSIS
[<function handler at 0x...>, <function handler at 0x...>]
>>> x.foo.handlers_full                                    # doctest: +ELLIPSIS
[(<function handler at 0x...>, (), {}), ...]
>>> x.foo.remove(handler=handler)                          # doctest: +ELLIPSIS
[...]

O último método, por exemplo, remove todas as respostas associadas ao sinal
"foo" implementadas pela função handler.

>>> x.trigger_foo()

É possível remover funções usando diversos critérios, mas o mais seguro é
utilizar o valor de saída do método `.listen()` que identifica cada função de
forma única. O método remove() sempre retorna uma lista com todos os handlers
removidos.

>>> rm_list = x.bar.remove(id=id2)
>>> x.trigger_bar()

Como podemos ver, a lista está vazia e x.trigger_bar() não produz nada. É
possível reintroduzir todos estes handlers usando o método .restore() do sinal.

>>> x.bar.restore(rm_list)
>>> x.trigger_bar()
called with args=(), kwargs={}


Até aqui, consideramos apenas funções chamadas sem argumentos para responder
aos sinais. Sinais podem ser declarados com um segundo argumento `num_args`
para definir o número de argumentos que devem ser passados para os métodos
trigger*.

>>> class Foobar(EventDispatcher):
...     foo = signal('foo', num_args=0)
...     bar = signal('bar', num_args=1)

Neste caso, o sinal "foo" continua funcionando como anteriormente, mas o sinal
"bar" precisa de um argumento adicional.

>>> x = Foobar()
>>> id1 = x.listen('bar', handler)

Agora o sinal "bar" deve ser disparado com um argumento posicional

>>> x.trigger_bar(0)
called with args=(0,), kwargs={}

Disparar o sinal com o número errado de argumentos provoca um TypeError

>>> x.trigger_bar()                         # doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
...
TypeError: trigger_method() missing 1 required positional argument: 'arg'

Caso se registre um método com argumentos posicionais, assumimos que estes
serão passados *após* os argumentos obrigatórios.

>>> id2 = x.listen('bar', handler, 1)
>>> x.trigger_bar(0)
called with args=(0,), kwargs={}
called with args=(0, 1), kwargs={}

Sinais filtrados
================

Até agora, o método trigger('sinal', *args) aceitam argumentos posicionais que
são passados diretamente para o argumento das funções do tipo "handler".
Sinais filtrados permitem passar argumentos para a função trigger que irão
determinar quais handlers serão chamados posteriormente.

Uma aplicação natural para essa idéia é o controle de eventos do teclado. Em
determinadas situações, um objeto pode querer ouvir todos os eventos de tecla
pressionada para, por exemplo, registrar o que o usuário digitou na tela.
Existem objetos, no entanto, que podem estar interessados apenas em algumas
combinações de teclas e desejam ouvir apenas teclas específicas. Sinais que
permitem fazer este tipo de distinção são aqui denominados sinais filtrados.

Considere o exemplo abaixo

>>> class Foobar(EventDispatcher):
...     foo = signal('key-up', 'key')
>>> x = Foobar()

Agora devemos registrar os sinais de foo com um argumento adicional 'key'

>>> def handler(*args, **kwds):
...     print('called with args=%s, kwargs=%s' % (args, kwds))
>>> id1 = x.listen('key-up', '<return>', handler)
>>> id2 = x.listen('key-up', '<space>', handler, 'spaced')

Podemos disparar o evento especificando o filtro a ser aplicado

>>> x.trigger_key_up('<return>')
called with args=(), kwargs={}
>>> x.trigger_key_up('<space>')
called with args=('spaced',), kwargs={}

Se não houver nenhum handler registrado para aquele filtro, então nada acontece

>>> x.trigger_key_up('other')

É possível definir "handlers" que atuam em todas as chamadas do sinal 'key-up',
independentemente do segundo argumento. Neste caso, o segundo argumento é
passado para o "handler" que decide então que fazer. Para isto, basta invocar
a função listen com None como filtro.

>>> id3 = x.listen('key-up', None, handler)
>>> x.trigger_key_up('<return>')
called with args=(), kwargs={}
called with args=('<return>',), kwargs={}

Caso a função não tenha handlers registrados, somente estes handlers genéricos
serão executados

>>> x.trigger_key_up('other')
called with args=('other',), kwargs={}

Quando registramos um novo handler, a ordem de execução é mantida como se os
handlers genéricos já tivessem sido registrados anteriormente para todas as
combinações de filtros possíveis.

>>> id4 = x.listen('key-up', 'other', handler, foo='bar')
>>> x.trigger_key_up('other')
called with args=('other',), kwargs={}
called with args=(), kwargs={'foo': 'bar'}

Subclasses
==========

Subclasses de EventDispatcher podem implementar métodos que são registrados
automaticamente pelo mecanismo listen() durante a inicialização dos objetos.
Para isto, basta utilizar a convenção de nomer os métodos como
`on_<nome do sinal>()`.

>>> class Foobar(EventDispatcher):
...     foo = signal('foo')
...     bar = signal('bar', 'key')
...
...     def on_foo(self):
...         print('foo triggered!')
...
...     def on_bar(self, key):
...         print('bar triggered with %s!' % key)

Deste modo, os métodos on_foo e on_bar já estão registrados pelo construtor da
classe Foobar para ouvir os respectivos sinais.

>>> x = Foobar()
>>> x.trigger_foo()
foo triggered!
>>> x.trigger_bar(0)
bar triggered with 0!

Uma maneira alternativa de atingir um resultado semelhante é utilizar o
decorador @listen. Deste modo, podemos atribuir handlers à sinais de forma
explicita no tempo de construção da classe. Este método é um pouco mais
flexível que o on_<sinal>() pois permite especifica filtros para sinais
filtrados.

>>> class Foobar(EventDispatcher):
...     foo = signal('foo')
...     bar = signal('bar', 'key')
...     foobar = signal('foobar', 'key')
...
...     @listen('foo')
...     def func1(self):
...         print('foo triggered!')
...
...     @listen('bar', 0)
...     @listen('bar', 1)
...     def func2(self):
...         print('bar triggered!')
...
...     @listen('foobar')
...     def func3(self, key):
...         print('foobar triggered with %s!' % key)



Agora o método func1() foi atribuído ao sinal 'foo' e o método func2() é
executado quando 'bar' é disparado com os valores 0 ou 1.

>>> x = Foobar()
>>> x.trigger_foo()
foo triggered!
>>> x.trigger_bar(0)
bar triggered!
>>> x.trigger_bar(2)
>>> x.trigger_foobar(3)
foobar triggered with 3!

Sinais delegados
================

As vezes é interessante expor um evento ao usuário mesmo quando a classe não
pode ser responsável por dispará-lo. Sinais delegados fazem justamente isso:
sempre que o usuário registrar um handler, ele delegará o registro para um
atributo específico da classe. Considere o exemplo mais concreto:

>>> class Foo(EventDispatcher):
...      do_foo = signal('do-foo', num_args=1)

A classe Foo é responsável por acionar os sinais 'do-foo'. Agora criamos uma
classe Bar que delega o sinal 'do-foo' para um atributo.

>>> class Bar(EventDispatcher):
...     def __init__(self, foo):
...         self._foo = foo
...
...     do_foo = signal('do-foo', delegate_to='_foo')

Criamos um objeto e registramos um handler

>>> foo = Foo()
>>> bar = Bar(foo)
>>> id1 = bar.listen('do-foo', handler)

Observe que bar não pode acionar diretamente o evento 'do-foo', mas temos que
fazer isto via o objeto originalmente responsável pelo evento

>>> bar.trigger_do_foo(0)
Traceback (most recent call last):
...
RuntimeError: signals can only be trigged by the owner

>>> foo.trigger_do_foo(0)
called with args=(0,), kwargs={}

A principal vantagem de delegar um sinal está no processo de destruição do
mesmo. Se chamarmos a função destroy_handlers() do objeto bar, todos os sinais
atribuídos a ele serão limpados, inclusive aqueles delegados para ``foo``. Isto
permite um controle mais granular sobre quem possui a referência sobre cada
sinal.pyname

>>> bar.destroy_handlers()                                 # doctest: +ELLIPSIS
{...}
>>> foo.do_foo.handlers
[]
'''

__package__ = 'FGAme.events'
if __name__ == '__main__':
    import FGAme


from .util import listen, signal
from .base import EventDispatcher, EventDispatcherMeta
from .dispatcher import dispatcher

if __name__ == '__main__':
    import doctest
    doctest.testmod()
    doctest.testmod(optionflags=doctest.REPORT_ONLY_FIRST_FAILURE)



# -*- coding: utf8 -*-

from FGAme import conf
from FGAme.util import lazy
from FGAme.mathtools import Vec2

__all__ = ['pos', 'vel']


def _factory_pos_property(pos):
    '''Cria propriedades que retornam as posições absolutas no objeto `pos`'''
    a, b = pos

    def pos_prop(self):
        w, h = self._shape
        x0, y0 = self._origin
        return Vec2(x0 + a * w, y0 + b * h)

    pos_prop.terms = (a, b)

    return property(pos_prop)


def _factory_from_func(prop):
    '''Cria funções que retornam posições relativas no objeto `pos`'''

    a, b = prop.fget.terms

    def from_pos(self, x, y):
        w, h = self._shape
        x0, y0 = self._origin
        return Vec2(x0 + a * w + x, y0 + b * h + y)

    return from_pos


class GlobalObject(object):

    '''Base para PosObject e VelObject'''

    @lazy
    def _globals(self):
        conf.init()
        self._globals = conf
        return self._globals

    @property
    def _shape(self):
        return self._globals.get_resolution()

    @property
    def _origin(self):
        return Vec2(0, 0)


class PosObject(GlobalObject):

    '''
    Implementa o objeto "pos", que permite acessar facilmente algumas posições
    na tela.

    Exemplos
    --------

    O objeto pos conhece as coordenadas de vários pontos de interesse na tela

    >>> pos.middle, pos.nw, pos.ne, pos.sw, pos.se
    (Vec(400, 300), Vec(0, 600), Vec(800, 600), Vec(0, 0), Vec(800, 0))

    >>> pos.north, pos.south, pos.east, pos.west
    (Vec(400, 600), Vec(400, 0), Vec(800, 300), Vec(0, 300))

    Também podemos especificar coordenadas relativas a partir de qualquer um
    destes pontos

    >>> pos.from_middle(100, 100), pos.from_ne(0, -100)
    (Vec(500, 400), Vec(800, 500))
    '''

    # Alinhados
    middle, north, south, east, west = \
        map(_factory_pos_property,
            [(.5, .5), (.5, 1), (.5, 0), (1, .5), (0, .5)])

    # Alinhados relativos
    from_middle, from_north, from_south, from_east, from_west = \
        map(_factory_from_func, [middle, north, south, east, west])

    # Diagonais
    sw, se, ne, nw = \
        map(_factory_pos_property, [(0, 0), (1, 0), (1, 1), (0, 1)])

    # Diagnoais (verboso)
    south_west, sout_east, north_east, nort_west = sw, se, ne, nw

    # Relativos diagonais
    from_sw, from_se, from_ne, from_nw = \
        map(_factory_from_func, [sw, se, ne, nw])

    # Relativos diagonais (verboso)
    from_south_west, from_south_east, from_north_east, from_north_west = \
        from_sw, from_se, from_ne, from_nw


class VelObject(GlobalObject):

    '''Implementa o objeto `vel`, que permite definir facilmente algumas
    velocidades no mundo'''

    def set_speeds(self, slow, fair, fast):
        '''Define o padrão de referência para as velocidades lenta, média e
        rápida'''

        if slow <= fair <= fast:
            self.set_slow(slow)
            self.set_fair(fair)
            self.set_fast(fast)

    def set_fast(self, value):
        '''Define a velocidade referência considerada "rápida"'''

        self._globals.speed_fast = value

    def set_fair(self, value):
        '''Define a velocidade referência considerada "média"'''

        self._globals.speed_fair = value

    def set_slow(self, value):
        '''Define a velocidade referência considerada "lenta"'''

        self._globals.speed_slow = value

    @property
    def fast(self):
        return self._globals.speed_fast

    @property
    def slow(self):
        return self._globals.speed_slow

    @property
    def fair(self):
        return self._globals.speed_fair

    # TODO: definir atributos e métodos para as velocidades
    # Velocidades aleatórias ##################################################
    def _random(self, scale, angle):
        return scale * vec_x.rotate(random.uniform(0, 2 * pi))

    def random(self, angle=None):
        return self._random(self.fair, angle)

    def random_fast(self, angle=None):
        return self._random(self.slow, angle)

    def random_slow(self, angle=None):
        return self._random(self.slow, angle)

# Velocidades em direções específicas #########################################
_speeds = dict(
    up=Vec2(0, 1),
    down=Vec2(0, -1),
    right=Vec2(1, 0),
    left=Vec2(-1, 0),
    ne=Vec2(1, 1).normalize(),
    nw=Vec2(-1, 1).normalize(),
    se=Vec2(1, -1).normalize(),
    sw=Vec2(-1, -1).normalize(),
)


def speed_prop(name, scale, Vec):
    def prop(self):
        return getattr(self, scale) * Vec
    prop.__name__ = name
    return property(prop)

for name, Vec in _speeds.items():
    setattr(VelObject, name, speed_prop(name, 'fair', Vec))
    setattr(VelObject, name, speed_prop('%s_fast' % name, 'fast', Vec))
    setattr(VelObject, name, speed_prop('%s_slow' % name, 'slow', Vec))

# Inicializa objetos
vel = VelObject()
pos = PosObject()

if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf8 -*-_

'''
Este exemplo demonstra a resposta a colisões com atrito utilizando duas caixas
AABB.
'''

from FGAme import World, AABB, Circle, RegularPoly, Rectangle, pi

stacks = []
stacks.append('aabbs')
stacks.append('rects')
stacks.append('circles')
stacks.append('polys')


# Cria mundo com coeficiente de atrito global não-nulo
world = World(friction=0.5, gravity=500, restitution=0.7, adamping=0.1)
world.add_bounds(width=10)

# Cria pilha de AABBs
if 'aabbs' in stacks:
    A = AABB(pos=(100, 70), shape=(100, 50), color='red', mass='inf')
    B = AABB(pos=(125, 150), shape=(50, 50))
    C = AABB(pos=(125, 250), shape=(70, 50))
    D = AABB(pos=(125, 300), shape=(100, 20))
    aabbs = [B, C, D]
    world.add([A, B, C, D])

# Cria pilha de Retângulos
if 'rects' in stacks:
    A = Rectangle(pos=(100, 370), shape=(100, 50), color='red', mass='inf')
    B = Rectangle(pos=(125, 450), shape=(50, 50))
    C = Rectangle(pos=(125, 510), shape=(70, 50))
    D = Rectangle(pos=(125, 550), shape=(100, 20))
    world.add([A, B, C, D])


# Cria pilha de círculos
if 'circles' in stacks:
    A = Circle(50, (250, 100), mass='inf', color='red')
    B = Circle(25, (320, 70), mass='inf', color='red')
    C = Circle(25, (380, 70), mass='inf', color='red')
    D = Circle(50, (450, 100), mass='inf', color='red')
    world.add([A, B, C, D])
    for i in range(11):
        c1 = Circle(8, pos=(300 + i * 10, 500 - i * 10))
        c2 = Circle(8, pos=(300 + i * 10, 200 + i * 10))
        world.add([c1, c2])

# Cria pilha de triângulos
if 'polys' in stacks:
    A = RegularPoly(3, 150, pos=(550, 70), theta=-pi / 6, mass='inf',
                    color='red')
    B = RegularPoly(3, 150, pos=(700, 70), theta=-pi / 6, mass='inf',
                    color='red')
    world.add([A, B])
    for i in range(10):
        o1 = RegularPoly(3, 20, pos=(550 + i * 12, 550 - i * 15))
        o2 = Rectangle(shape=(12, 12), pos=(550 + i * 10, 200 + i * 10))
        world.add([o1, o2])


#@world.listen('frame-enter')
# def show_vels():
#    print('1) %7.1f, 2) %7.1f, 3) %7.1f' % tuple(obj.vel.y for obj in aabbs))


# Inicia a simulação
world.run()
